# SPDX-License-Identifier: LGPL-2.1-or-later

import json

import pytest
import yaml

import libnmstate
from libnmstate.schema import DNS
from libnmstate.schema import Interface
from libnmstate.schema import InterfaceIPv4
from libnmstate.schema import InterfaceIPv6
from libnmstate.schema import InterfaceState
from libnmstate.schema import InterfaceType
from libnmstate.schema import Route

from .testlib import assertlib
from .testlib import cmdlib
from .testlib.bondlib import bond_interface
from .testlib.genconf import gen_conf_apply

IPV4_DNS_NAMESERVERS = ["8.8.8.8", "1.1.1.1"]
EXTRA_IPV4_DNS_NAMESERVER = "9.9.9.9"
IPV6_DNS_NAMESERVERS = ["2001:4860:4860::8888", "2606:4700:4700::1111"]
IPV6_DNS_LONG_NAMESERVER = ["2000:0000:0000:0000:0000:0000:0000:0100"]
EXTRA_IPV6_DNS_NAMESERVER = "2620:fe::9"
EXAMPLE_SEARCHES = ["example.org", "example.com"]
EXAMPLE_SEARCHES2 = ["example.info", "example.org"]
TEST_BOND0 = "test-bond0"

parametrize_ip_ver = pytest.mark.parametrize(
    "dns_config",
    [
        ({DNS.SERVER: IPV4_DNS_NAMESERVERS, DNS.SEARCH: EXAMPLE_SEARCHES}),
        ({DNS.SERVER: IPV6_DNS_NAMESERVERS, DNS.SEARCH: EXAMPLE_SEARCHES}),
    ],
    ids=["ipv4", "ipv6"],
)

DUMMY0 = "dummy0"


@pytest.fixture(scope="function", autouse=True)
def dns_test_env(eth1_up, eth2_up):
    yield
    # Remove DNS config as it be saved in eth1 or eth2 which might trigger
    # failure when bring eth1/eth2 down.
    desired_state = {
        Interface.KEY: _get_test_iface_states(),
        DNS.KEY: {DNS.CONFIG: {DNS.SERVER: [], DNS.SEARCH: []}},
    }
    libnmstate.apply(desired_state)


@pytest.mark.tier1
@parametrize_ip_ver
def test_dns_edit_nameserver_with_static_gateway(dns_config):
    desired_state = {
        Interface.KEY: _get_test_iface_states(),
        Route.KEY: {Route.CONFIG: _gen_default_gateway_route()},
        DNS.KEY: {DNS.CONFIG: dns_config},
    }
    libnmstate.apply(desired_state)
    current_state = libnmstate.show()
    assert dns_config == current_state[DNS.KEY][DNS.CONFIG]


def test_dns_edit_ipv4_nameserver_before_ipv6():
    dns_config = {
        DNS.SERVER: [IPV4_DNS_NAMESERVERS[0], IPV6_DNS_NAMESERVERS[0]],
        DNS.SEARCH: [],
    }
    desired_state = {
        Interface.KEY: _get_test_iface_states(),
        Route.KEY: {Route.CONFIG: _gen_default_gateway_route()},
        DNS.KEY: {DNS.CONFIG: dns_config},
    }
    libnmstate.apply(desired_state)
    current_state = libnmstate.show()
    assert dns_config == current_state[DNS.KEY][DNS.CONFIG]


@pytest.mark.tier1
def test_dns_edit_ipv6_nameserver_before_ipv4():
    dns_config = {
        DNS.SERVER: [IPV6_DNS_NAMESERVERS[0], IPV4_DNS_NAMESERVERS[0]],
        DNS.SEARCH: [],
    }
    desired_state = {
        Interface.KEY: _get_test_iface_states(),
        Route.KEY: {Route.CONFIG: _gen_default_gateway_route()},
        DNS.KEY: {DNS.CONFIG: dns_config},
    }
    libnmstate.apply(desired_state)
    current_state = libnmstate.show()
    assert dns_config == current_state[DNS.KEY][DNS.CONFIG]


@pytest.mark.tier1
@pytest.mark.parametrize(
    "dns_servers",
    [
        (IPV4_DNS_NAMESERVERS + [EXTRA_IPV4_DNS_NAMESERVER]),
        (IPV6_DNS_NAMESERVERS + [EXTRA_IPV6_DNS_NAMESERVER]),
        (IPV4_DNS_NAMESERVERS + [EXTRA_IPV6_DNS_NAMESERVER]),
        (IPV6_DNS_NAMESERVERS + [EXTRA_IPV4_DNS_NAMESERVER]),
        [
            IPV4_DNS_NAMESERVERS[0],
            EXTRA_IPV6_DNS_NAMESERVER,
            IPV4_DNS_NAMESERVERS[1],
        ],
        [
            IPV6_DNS_NAMESERVERS[0],
            EXTRA_IPV4_DNS_NAMESERVER,
            IPV6_DNS_NAMESERVERS[1],
        ],
        (IPV4_DNS_NAMESERVERS + IPV6_DNS_NAMESERVERS),
        (IPV6_DNS_NAMESERVERS + IPV4_DNS_NAMESERVERS),
    ],
    ids=[
        "3ipv4",
        "3ipv6",
        "2ipv4+ipv6",
        "2ipv6+ipv4",
        "ipv4+ipv6+ipv4",
        "ipv6+ipv4+ipv6",
        "2ipv4+2ipv6",
        "2ipv6+2ipv4",
    ],
)
def test_dns_edit_3_more_nameservers(dns_servers):
    dns_config = {
        DNS.SERVER: dns_servers,
        DNS.SEARCH: [],
    }
    desired_state = {
        Interface.KEY: _get_test_iface_states(),
        Route.KEY: {Route.CONFIG: _gen_default_gateway_route()},
        DNS.KEY: {DNS.CONFIG: dns_config},
    }
    libnmstate.apply(desired_state)
    current_state = libnmstate.show()
    assert dns_config == current_state[DNS.KEY][DNS.CONFIG]


@pytest.mark.tier1
@pytest.mark.parametrize(
    "empty_dns_config",
    [{DNS.SERVER: [], DNS.SEARCH: []}, {}],
    ids=[
        "empty_server_and_search",
        "empty_dict",
    ],
)
def test_remove_dns_config(empty_dns_config):
    dns_config = {
        DNS.SERVER: [IPV6_DNS_NAMESERVERS[0], IPV4_DNS_NAMESERVERS[0]],
        DNS.SEARCH: [],
    }
    desired_state = {
        Interface.KEY: _get_test_iface_states(),
        Route.KEY: {Route.CONFIG: _gen_default_gateway_route()},
        DNS.KEY: {DNS.CONFIG: dns_config},
    }
    libnmstate.apply(desired_state)

    libnmstate.apply(
        {Interface.KEY: [], DNS.KEY: {DNS.CONFIG: empty_dns_config}}
    )
    current_state = libnmstate.show()
    assert {} == current_state[DNS.KEY][DNS.CONFIG]


@pytest.fixture
def dummy0_up():
    dummy_iface_state = {
        Interface.NAME: DUMMY0,
        Interface.TYPE: InterfaceType.DUMMY,
        Interface.STATE: InterfaceState.UP,
    }
    libnmstate.apply({Interface.KEY: [dummy_iface_state]})
    yield dummy_iface_state
    dummy_iface_state[Interface.STATE] = InterfaceState.ABSENT
    libnmstate.apply({Interface.KEY: [dummy_iface_state]})


def test_preserve_dns_config(dummy0_up):
    dns_config = {
        DNS.SERVER: [IPV6_DNS_NAMESERVERS[0], IPV4_DNS_NAMESERVERS[0]],
        DNS.SEARCH: [],
    }
    desired_state = {
        Interface.KEY: _get_test_iface_states(),
        Route.KEY: {Route.CONFIG: _gen_default_gateway_route()},
        DNS.KEY: {DNS.CONFIG: dns_config},
    }
    libnmstate.apply(desired_state)

    # Add new dummy interface with default gateway, nmstate should
    # preserve the existing DNS configure as interface holding DNS
    # configuration is not changed and DNS configure is still the same.
    libnmstate.apply(
        {
            Interface.KEY: [
                {
                    Interface.NAME: DUMMY0,
                    Interface.TYPE: InterfaceType.DUMMY,
                    Interface.IPV4: {
                        InterfaceIPv4.ADDRESS: [
                            {
                                InterfaceIPv4.ADDRESS_IP: "192.0.2.250",
                                InterfaceIPv4.ADDRESS_PREFIX_LENGTH: 24,
                            }
                        ],
                        InterfaceIPv4.ENABLED: True,
                    },
                    Interface.IPV6: {
                        InterfaceIPv6.ADDRESS: [
                            {
                                InterfaceIPv6.ADDRESS_IP: "2001:db8:f::1",
                                InterfaceIPv6.ADDRESS_PREFIX_LENGTH: 64,
                            }
                        ],
                        InterfaceIPv6.ENABLED: True,
                    },
                }
            ],
            Route.KEY: {
                Route.CONFIG: [
                    {
                        Route.DESTINATION: "0.0.0.0/0",
                        Route.NEXT_HOP_ADDRESS: "192.0.2.2",
                        Route.NEXT_HOP_INTERFACE: DUMMY0,
                    },
                    {
                        Route.DESTINATION: "::/0",
                        Route.NEXT_HOP_ADDRESS: "2001:db8:f::2",
                        Route.NEXT_HOP_INTERFACE: DUMMY0,
                    },
                ]
            },
        }
    )

    current_state = libnmstate.show()

    assert dns_config == current_state[DNS.KEY][DNS.CONFIG]


@pytest.fixture
def setup_ipv4_ipv6_name_server():
    desired_state = {
        Interface.KEY: _get_test_iface_states(),
        Route.KEY: {Route.CONFIG: _gen_default_gateway_route()},
        DNS.KEY: {
            DNS.CONFIG: {
                DNS.SERVER: [IPV6_DNS_NAMESERVERS[0], IPV4_DNS_NAMESERVERS[0]],
                DNS.SEARCH: [],
            }
        },
    }
    libnmstate.apply(desired_state)
    yield desired_state


def test_preserve_dns_config_with_empty_state(setup_ipv4_ipv6_name_server):
    old_state = setup_ipv4_ipv6_name_server

    libnmstate.apply({Interface.KEY: []})
    current_state = libnmstate.show()

    assert old_state[DNS.KEY][DNS.CONFIG] == current_state[DNS.KEY][DNS.CONFIG]


def test_add_non_canonicalized_ipv6_nameserver():
    dns_config = {
        DNS.SERVER: IPV6_DNS_LONG_NAMESERVER,
        DNS.SEARCH: [],
    }
    desired_state = {
        Interface.KEY: _get_test_iface_states(),
        Route.KEY: {Route.CONFIG: _gen_default_gateway_route()},
        DNS.KEY: {
            DNS.CONFIG: dns_config,
        },
    }
    libnmstate.apply(desired_state)

    current_state = libnmstate.show()
    assert "2000::100" in current_state[DNS.KEY][DNS.CONFIG][DNS.SERVER]


def _get_test_iface_states():
    return [
        {
            Interface.NAME: "eth1",
            Interface.STATE: InterfaceState.UP,
            Interface.TYPE: InterfaceType.ETHERNET,
            Interface.IPV4: {
                InterfaceIPv4.ADDRESS: [
                    {
                        InterfaceIPv4.ADDRESS_IP: "192.0.2.251",
                        InterfaceIPv4.ADDRESS_PREFIX_LENGTH: 24,
                    }
                ],
                InterfaceIPv4.DHCP: False,
                InterfaceIPv4.ENABLED: True,
            },
            Interface.IPV6: {
                InterfaceIPv6.ADDRESS: [
                    {
                        InterfaceIPv6.ADDRESS_IP: "2001:db8:1::1",
                        InterfaceIPv6.ADDRESS_PREFIX_LENGTH: 64,
                    }
                ],
                InterfaceIPv6.DHCP: False,
                InterfaceIPv6.AUTOCONF: False,
                InterfaceIPv6.ENABLED: True,
            },
        },
        {
            Interface.NAME: "eth2",
            Interface.STATE: InterfaceState.UP,
            Interface.TYPE: InterfaceType.ETHERNET,
            Interface.IPV4: {
                InterfaceIPv4.ADDRESS: [
                    {
                        InterfaceIPv4.ADDRESS_IP: "198.51.100.1",
                        InterfaceIPv4.ADDRESS_PREFIX_LENGTH: 24,
                    }
                ],
                InterfaceIPv4.DHCP: False,
                InterfaceIPv4.ENABLED: True,
            },
            Interface.IPV6: {
                InterfaceIPv6.ADDRESS: [
                    {
                        InterfaceIPv6.ADDRESS_IP: "2001:db8:2::1",
                        InterfaceIPv6.ADDRESS_PREFIX_LENGTH: 64,
                    }
                ],
                InterfaceIPv6.DHCP: False,
                InterfaceIPv6.AUTOCONF: False,
                InterfaceIPv6.ENABLED: True,
            },
        },
    ]


def _gen_default_gateway_route():
    return [
        {
            Route.DESTINATION: "0.0.0.0/0",
            Route.METRIC: 200,
            Route.NEXT_HOP_ADDRESS: "192.0.2.1",
            Route.NEXT_HOP_INTERFACE: "eth1",
        },
        {
            Route.DESTINATION: "::/0",
            Route.METRIC: 201,
            Route.NEXT_HOP_ADDRESS: "2001:db8:2::f",
            Route.NEXT_HOP_INTERFACE: "eth1",
        },
    ]


@pytest.fixture
def static_dns(eth1_up):
    desired_state = {
        Interface.KEY: _get_test_iface_states(),
        Route.KEY: {Route.CONFIG: _gen_default_gateway_route()},
        DNS.KEY: {
            DNS.CONFIG: {
                DNS.SERVER: [IPV6_DNS_NAMESERVERS[0], IPV4_DNS_NAMESERVERS[0]],
                DNS.SEARCH: EXAMPLE_SEARCHES,
            }
        },
    }
    libnmstate.apply(desired_state)
    yield desired_state
    libnmstate.apply({DNS.KEY: {DNS.CONFIG: {}}})


@pytest.mark.tier1
def test_change_dns_search_only(static_dns):
    desired_state = {
        DNS.KEY: {DNS.CONFIG: {DNS.SEARCH: EXAMPLE_SEARCHES2}},
    }
    libnmstate.apply(desired_state)
    current_state = libnmstate.show()
    assert current_state[DNS.KEY][DNS.CONFIG] == {
        DNS.SERVER: [IPV6_DNS_NAMESERVERS[0], IPV4_DNS_NAMESERVERS[0]],
        DNS.SEARCH: EXAMPLE_SEARCHES2,
    }


def test_change_dns_server_only(static_dns):
    desired_state = {
        DNS.KEY: {
            DNS.CONFIG: {
                DNS.SERVER: [IPV6_DNS_NAMESERVERS[1], IPV4_DNS_NAMESERVERS[1]]
            }
        },
    }
    libnmstate.apply(desired_state)
    current_state = libnmstate.show()
    assert current_state[DNS.KEY][DNS.CONFIG] == {
        DNS.SERVER: [IPV6_DNS_NAMESERVERS[1], IPV4_DNS_NAMESERVERS[1]],
        DNS.SEARCH: EXAMPLE_SEARCHES,
    }


def test_nmstatectl_show_dns(static_dns):
    rc, out, err = cmdlib.exec_cmd("nmstatectl show --json".split())
    assert rc == 0
    current_state = json.loads(out)
    assert (
        current_state[DNS.KEY][DNS.CONFIG] == static_dns[DNS.KEY][DNS.CONFIG]
    )


@pytest.mark.tier1
@parametrize_ip_ver
def test_dns_edit_nameserver_with_static_gateway_genconf(dns_config):
    desired_state = {
        Interface.KEY: _get_test_iface_states(),
        Route.KEY: {Route.CONFIG: _gen_default_gateway_route()},
        DNS.KEY: {DNS.CONFIG: dns_config},
    }
    with gen_conf_apply(desired_state):
        current_state = libnmstate.show()
        assert dns_config == current_state[DNS.KEY][DNS.CONFIG]


def test_move_dns_from_port_to_controller(static_dns, eth2_up):
    with bond_interface(
        name=TEST_BOND0, port=["eth1", "eth2"], create=False
    ) as state:
        state[Route.KEY] = static_dns[Route.KEY]
        for route in state[Route.KEY][Route.CONFIG]:
            route[Route.NEXT_HOP_INTERFACE] = TEST_BOND0

        state[DNS.KEY] = static_dns[DNS.KEY]
        state[Interface.KEY][0][Interface.IPV4] = static_dns[Interface.KEY][0][
            Interface.IPV4
        ]
        state[Interface.KEY][0][Interface.IPV6] = static_dns[Interface.KEY][0][
            Interface.IPV6
        ]
        libnmstate.apply(state)
        current_state = libnmstate.show()

        assertlib.assert_state_match(state)
        assert state[DNS.KEY][DNS.CONFIG] == current_state[DNS.KEY][DNS.CONFIG]
        # Remove DNS before deleting bond
        libnmstate.apply({DNS.KEY: {DNS.CONFIG: {}}})


def test_changed_dns_from_port_to_controller(static_dns, eth2_up):
    with bond_interface(
        name=TEST_BOND0, port=["eth1", "eth2"], create=False
    ) as state:
        state[Route.KEY] = static_dns[Route.KEY]
        for route in state[Route.KEY][Route.CONFIG]:
            route[Route.NEXT_HOP_INTERFACE] = TEST_BOND0

        state[DNS.KEY] = static_dns[DNS.KEY]
        state[DNS.KEY][DNS.CONFIG][DNS.SEARCH].reverse()
        state[DNS.KEY][DNS.CONFIG][DNS.SERVER].reverse()
        state[Interface.KEY][0][Interface.IPV4] = static_dns[Interface.KEY][0][
            Interface.IPV4
        ]
        state[Interface.KEY][0][Interface.IPV6] = static_dns[Interface.KEY][0][
            Interface.IPV6
        ]
        libnmstate.apply(state)
        current_state = libnmstate.show()

        assertlib.assert_state_match(state)
        assert state[DNS.KEY][DNS.CONFIG] == current_state[DNS.KEY][DNS.CONFIG]
        # Remove DNS before deleting bond
        libnmstate.apply({DNS.KEY: {DNS.CONFIG: {}}})


def test_uncompare_dns_servers(static_dns):
    desired_state = yaml.load(
        """---
        dns-resolver:
          config:
            server:
            - 2001:Db8:0:0:0:0:0:1
            - ::fFfF:192.0.2.1
        """,
        Loader=yaml.SafeLoader,
    )
    libnmstate.apply(desired_state)
    current_state = libnmstate.show()
    assert "2001:db8::1" in current_state[DNS.KEY][DNS.CONFIG][DNS.SERVER]
    assert "::ffff:192.0.2.1" in current_state[DNS.KEY][DNS.CONFIG][DNS.SERVER]



#
# Copyright (c) 2021 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from contextlib import contextmanager
import os
import tempfile


@contextmanager
def tmp_plugin_dir():
    with tempfile.TemporaryDirectory() as plugin_dir:
        os.environ["NMSTATE_PLUGIN_DIR"] = plugin_dir
        try:
            yield plugin_dir
        finally:
            os.environ.pop("NMSTATE_PLUGIN_DIR")



# SPDX-License-Identifier: Apache-2.0


import json

from .clib_wrapper import net_state_from_policy


def gen_net_state_from_policy(policy, cur_state):
    return json.loads(net_state_from_policy(policy, cur_state))



#
# Copyright (c) 2022 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

import os

import pytest

import libnmstate
from libnmstate.error import NmstateVerificationError
from libnmstate.schema import HostNameState

from .testlib import cmdlib

TEST_HOSTNAME1 = "nmstate-test1.example.org"
TEST_HOSTNAME2 = "nmstate-test2.example.org"


@pytest.fixture
def restore_hostname():
    cur_hostname_conf = libnmstate.show()[HostNameState.KEY]
    yield
    libnmstate.apply({HostNameState.KEY: cur_hostname_conf})


@pytest.mark.tier1
@pytest.mark.skipif(
    os.environ.get("CI") == "true",
    reason="NM cannot change hostname in container",
)
def test_hostname_set_chg_and_clear(restore_hostname):
    libnmstate.apply(
        {
            HostNameState.KEY: {
                HostNameState.CONFIG: TEST_HOSTNAME1,
            }
        }
    )
    cur_host_name = cmdlib.exec_cmd(["hostname"], check=True)[1]
    assert os.path.exists("/etc/hostname")
    assert cur_host_name.strip() == TEST_HOSTNAME1
    libnmstate.apply(
        {
            HostNameState.KEY: {
                HostNameState.RUNNING: TEST_HOSTNAME2,
                HostNameState.CONFIG: "",
            }
        }
    )
    cur_host_name = cmdlib.exec_cmd(["hostname"], check=True)[1]
    assert cur_host_name.strip() == TEST_HOSTNAME2
    assert not os.path.exists("/etc/hostname")


@pytest.mark.skipif(
    os.environ.get("CI") == "true",
    reason="NM cannot change hostname in container",
)
def test_hostname_set_config_in_memory_only(restore_hostname):
    with pytest.raises(NmstateVerificationError):
        libnmstate.apply(
            {
                HostNameState.KEY: {
                    HostNameState.RUNNING: TEST_HOSTNAME2,
                    HostNameState.CONFIG: TEST_HOSTNAME2,
                }
            },
            save_to_disk=False,
        )


@pytest.mark.skipif(
    os.environ.get("CI") == "true",
    reason="NM cannot change hostname in container",
)
def test_hostname_set_in_memory_only(restore_hostname):
    libnmstate.apply(
        {
            HostNameState.KEY: {
                HostNameState.RUNNING: TEST_HOSTNAME2,
            }
        },
    )
    cur_host_name = cmdlib.exec_cmd(["hostname"], check=True)[1]
    assert cur_host_name.strip() == TEST_HOSTNAME2


@pytest.mark.skipif(
    os.environ.get("CI") == "true",
    reason="NM cannot change hostname in container",
)
def test_hostname_set_different_running_and_config(restore_hostname):
    libnmstate.apply(
        {
            HostNameState.KEY: {
                HostNameState.RUNNING: TEST_HOSTNAME1,
                HostNameState.CONFIG: TEST_HOSTNAME2,
            }
        },
    )
    cur_host_name = cmdlib.exec_cmd(["hostname"], check=True)[1]
    assert cur_host_name.strip() == TEST_HOSTNAME1
    assert (
        cmdlib.exec_cmd(["cat", "/etc/hostname"], check=True)[1].strip()
        == TEST_HOSTNAME2
    )



#
# Copyright (c) 2019-2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from contextlib import contextmanager
import copy

import libnmstate
from libnmstate.schema import Interface
from libnmstate.schema import InterfaceState
from libnmstate.schema import InterfaceType
from libnmstate.schema import OVSBridge
from libnmstate.schema import OVSInterface


class Bridge:
    def __init__(self, name):
        self._name = name
        self._ifaces = [
            {
                Interface.NAME: name,
                Interface.TYPE: InterfaceType.OVS_BRIDGE,
                OVSBridge.CONFIG_SUBTREE: {},
            }
        ]
        self._bridge_iface = self._ifaces[0]

    def set_options(self, options):
        self._bridge_iface[OVSBridge.CONFIG_SUBTREE][
            OVSBridge.OPTIONS_SUBTREE
        ] = options

    def set_ovs_db(self, ovs_db_config):
        self._bridge_iface[OVSBridge.OVS_DB_SUBTREE] = ovs_db_config

    def add_system_port(self, name):
        self._add_port(name)

    def add_link_aggregation_port(
        self, name, port, mode=None, updelay=None, downdelay=None
    ):
        self._add_port(name)
        port_cfg = self._get_port(name)
        port_cfg[OVSBridge.Port.LINK_AGGREGATION_SUBTREE] = {
            OVSBridge.Port.LinkAggregation.PORT_SUBTREE: [
                {OVSBridge.Port.LinkAggregation.Port.NAME: port_member}
                for port_member in port
            ]
        }
        if mode:
            port_cfg[OVSBridge.Port.LINK_AGGREGATION_SUBTREE][
                OVSBridge.Port.LinkAggregation.MODE
            ] = mode
        if updelay is not None:
            port_cfg[OVSBridge.Port.LINK_AGGREGATION_SUBTREE][
                OVSBridge.Port.LinkAggregation.Options.UP_DELAY
            ] = updelay
        if downdelay is not None:
            port_cfg[OVSBridge.Port.LINK_AGGREGATION_SUBTREE][
                OVSBridge.Port.LinkAggregation.Options.DOWN_DELAY
            ] = downdelay

    def add_internal_port(
        self,
        name,
        *,
        mac=None,
        copy_mac_from=None,
        ipv4_state=None,
        ovs_db=None,
        patch_state=None,
        mtu=None,
        dpdk_state=None,
    ):
        ifstate = {
            Interface.NAME: name,
            Interface.TYPE: InterfaceType.OVS_INTERFACE,
        }
        if mac:
            ifstate[Interface.MAC] = mac
        if copy_mac_from:
            ifstate[Interface.COPY_MAC_FROM] = copy_mac_from
        if mtu:
            ifstate[Interface.MTU] = mtu
        if ipv4_state:
            ifstate[Interface.IPV4] = ipv4_state
        if ovs_db:
            ifstate[OVSInterface.OVS_DB_SUBTREE] = ovs_db
        if patch_state:
            ifstate[OVSInterface.PATCH_CONFIG_SUBTREE] = patch_state
        if dpdk_state:
            ifstate[OVSInterface.DPDK_CONFIG_SUBTREE] = dpdk_state

        self._add_port(name)
        self._ifaces.append(ifstate)

    @property
    def ports_names(self):
        return [port[OVSBridge.Port.NAME] for port in self._get_ports()]

    def _add_port(self, name):
        self._bridge_iface[OVSBridge.CONFIG_SUBTREE].setdefault(
            OVSBridge.PORT_SUBTREE, []
        ).append({OVSBridge.Port.NAME: name})

    def _get_port(self, name):
        ports = self._get_ports()
        return next(
            (port for port in ports if port[OVSBridge.Port.NAME] == name), None
        )

    def _get_ports(self):
        return self._bridge_iface[OVSBridge.CONFIG_SUBTREE].get(
            OVSBridge.PORT_SUBTREE, []
        )

    def set_port_option(self, port_name, new_option):
        for port_option in self._bridge_iface[OVSBridge.CONFIG_SUBTREE].get(
            OVSBridge.PORT_SUBTREE, []
        ):
            if port_option[OVSBridge.Port.NAME] == port_name:
                port_option.update(new_option)

    @contextmanager
    def create(self):
        desired_state = self.state
        libnmstate.apply(desired_state)
        try:
            yield desired_state
        finally:
            desired_state = {
                Interface.KEY: _set_ifaces_state(
                    self._ifaces, InterfaceState.ABSENT
                )
            }
            libnmstate.apply(desired_state, verify_change=False)

    def apply(self):
        libnmstate.apply(self.state)

    @property
    def state(self):
        return {
            Interface.KEY: _set_ifaces_state(self._ifaces, InterfaceState.UP)
        }


def _set_ifaces_state(ifaces, state):
    ifaces = copy.deepcopy(ifaces)
    for iface in ifaces:
        iface[Interface.STATE] = state
    return ifaces


@contextmanager
def ovs_bridge(br_name, sys_port_names, internal_port_name):
    bridge = Bridge(br_name)
    bridge.add_internal_port(internal_port_name)
    for sys_port_name in sys_port_names:
        bridge.add_system_port(sys_port_name)
    with bridge.create():
        yield bridge.state


@contextmanager
# The bond_ports should be in the format of
#   dict(bond_port_name, bond_sys_port_names)
def ovs_bridge_bond(br_name, bond_ports, internal_port_name):
    bridge = Bridge(br_name)
    bridge.add_internal_port(internal_port_name)
    for bond_port_name, bond_sys_port_names in bond_ports.items():
        bridge.add_link_aggregation_port(bond_port_name, bond_sys_port_names)
    with bridge.create():
        yield bridge.state



# SPDX-License-Identifier: LGPL-2.1-or-later

import os

import gi

from .cmdlib import exec_cmd

gi.require_version("NM", "1.0")
# It is required to state the NM version before importing it
# But this break the flak8 rule: https://www.flake8rules.com/rules/E402.html
# Use NOQA: E402 to suppress it.
# pylint: disable=no-name-in-module
from gi.repository import NM  # NOQA: E402

# pylint: enable=no-name-in-module


def is_fedora():
    return os.path.exists("/etc/fedora-release")


def is_ubuntu_kernel():
    return "Ubuntu" in os.uname().version


def nm_major_minor_version():
    return float(f"{NM.MAJOR_VERSION}.{NM.MINOR_VERSION}")


def nm_minor_version():
    return int(f"{NM.MINOR_VERSION}")


def is_k8s():
    return os.getenv("RUN_K8S") == "true"


def is_el8():
    return exec_cmd("rpm -E %{?rhel}".split())[1].strip() == "8"



#
# Copyright (c) 2018-2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
from contextlib import contextmanager

from libnmstate import error

from ..testlib import cmdlib


class MainloopTestError(Exception):
    pass


@contextmanager
def main_context(ctx):
    yield
    try:
        ctx.wait_all_finish()
    except error.NmstateLibnmError as ex:
        raise MainloopTestError(str(ex.args))


def get_nm_active_profiles():
    all_profile_names_output = cmdlib.exec_cmd(
        "nmcli -g NAME connection show --active".split(" "), check=True
    )[1]
    all_profile_uuids_output = cmdlib.exec_cmd(
        "nmcli -g UUID connection show --active".split(" "), check=True
    )[1]
    return (
        all_profile_names_output.split("\n"),
        all_profile_uuids_output.split("\n"),
    )


def get_proxy_port_profile_uuid_of_ovs_interface(iface_name):
    proxy_port_uuid = cmdlib.exec_cmd(
        f"nmcli -g connection.master connection show {iface_name}".split(" "),
        check=True,
    )[1].strip()
    cmdlib.exec_cmd(
        f"nmcli -g connection.id connection show {proxy_port_uuid}".split(" "),
        check=True,
    )
    return proxy_port_uuid



#
# Copyright (c) 2018-2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

import logging
import os
import subprocess
import tempfile

import pytest

import libnmstate
from libnmstate.schema import DNS
from libnmstate.schema import Route
from libnmstate.schema import RouteRule

from .testlib import ifacelib


REPORT_HEADER = """RPMs: {rpms}
OS: {osname}
nmstate: {nmstate_version}
"""


def pytest_configure(config):
    config.addinivalue_line("markers", "slow: mark time consuming test")
    config.addinivalue_line("markers", "tier2")
    config.addinivalue_line("markers", "tier1")


def pytest_addoption(parser):
    parser.addoption(
        "--runslow", action="store_true", default=False, help="run slow tests"
    )


def pytest_collection_modifyitems(config, items):
    if not config.getoption("--runslow"):
        # --runslow is not in cli: skip slow tests
        _mark_skip_slow_tests(items)
    _mark_tier2_tests(items)


def _mark_skip_slow_tests(items):
    skip_slow = pytest.mark.skip(reason="need --runslow option to run")
    for item in items:
        if "slow" in item.keywords:
            item.add_marker(skip_slow)


def _mark_tier2_tests(items):
    for item in items:
        if "tier1" not in item.keywords:
            item.add_marker(pytest.mark.tier2)


@pytest.fixture(scope="session", autouse=True)
def fix_ip_netns_issue(scope="session", autouse=True):
    if os.getenv("CI"):
        with tempfile.TemporaryDirectory() as tmpdirname:
            subprocess.run(
                f"mount -t sysfs --make-private {tmpdirname}".split(),
                check=True,
            )
            yield
            subprocess.run(f"umount {tmpdirname}".split())
    else:
        yield


@pytest.fixture(scope="session", autouse=True)
def test_env_setup():
    _logging_setup()
    old_state = libnmstate.show()
    old_state = _remove_interfaces_from_env(old_state)
    _remove_dns_route_route_rule()
    _ethx_init()
    yield
    restore_old_state(old_state)


def _remove_dns_route_route_rule():
    """
    Remove existing DNS, routes, route rules in case it interference tests.
    """
    libnmstate.apply(
        {
            DNS.KEY: {DNS.CONFIG: {}},
            Route.KEY: {
                Route.CONFIG: [{Route.STATE: Route.STATE_ABSENT}],
            },
            RouteRule.KEY: {RouteRule.CONFIG: []},
        },
        verify_change=False,
    )


def _logging_setup():
    logging.basicConfig(
        format="%(asctime)s %(name)-12s %(levelname)-8s %(message)s",
        level=logging.DEBUG,
    )


def _ethx_init():
    """Remove any existing definitions on the ethX interfaces."""
    ifacelib.ifaces_init("eth1", "eth2")


def _remove_interfaces_from_env(state):
    """
    Remove references from interfaces passed to environment variable
    NMSTATE_TEST_IGNORE_IFACE.
    """
    ignore_iface = os.getenv("NMSTATE_TEST_IGNORE_IFACE")
    if ignore_iface is None:
        return state

    state["interfaces"] = [
        i for i in state["interfaces"] if ignore_iface not in i["name"]
    ]
    state["routes"]["config"] = [
        r
        for r in state["routes"]["config"]
        if ignore_iface not in r["next-hop-interface"]
    ]
    return state


@pytest.fixture(scope="function")
def eth1_up():
    with ifacelib.iface_up("eth1") as ifstate:
        yield ifstate


@pytest.fixture(scope="function")
def eth2_up():
    with ifacelib.iface_up("eth2") as ifstate:
        yield ifstate


port0_up = eth1_up
port1_up = eth2_up


def pytest_report_header(config):
    return REPORT_HEADER.format(
        rpms=_get_package_nvr("NetworkManager"),
        osname=_get_osname(),
        nmstate_version=_get_nmstate_version(),
    )


def _get_nmstate_version():
    """
    Prefer RPM version of nmstate, if not found, use libnmstate module version
    """
    try:
        return _get_package_nvr("nmstate")
    except subprocess.CalledProcessError:
        return libnmstate.__version__


def _get_package_nvr(package):
    return (
        subprocess.check_output(["rpm", "-q", package]).strip().decode("utf-8")
    )


def _get_osname():
    with open("/etc/os-release") as os_release:
        for line in os_release.readlines():
            if line.startswith("PRETTY_NAME="):
                return line.split("=", maxsplit=1)[1].strip().strip('"')
    return ""


# Only restore the interface with IPv4/IPv6 gateway with IP/DNS config only
# For test machine, it is expected to lose configurations
def restore_old_state(old_state):
    gw_routes = [
        rt
        for rt in old_state["routes"].get("config", [])
        if rt["destination"] in ("0.0.0.0/0", "::/0")
    ]
    gw_ifaces = [rt["next-hop-interface"] for rt in gw_routes]
    desire_state = {
        "interfaces": [],
        "routes": {"config": gw_routes},
        "dns-resolver": old_state.get("dns-resolver", {}),
    }
    for iface_name in gw_ifaces:
        for iface in old_state["interfaces"]:
            if iface["name"] in gw_ifaces:
                if iface["state"] == "up":
                    desire_state["interfaces"].append(
                        {
                            "name": iface["name"],
                            "type": iface["type"],
                            "ipv4": iface["ipv4"],
                            "ipv6": iface["ipv6"],
                        }
                    )
    if len(desire_state["interfaces"]):
        libnmstate.apply(desire_state, verify_change=False)



#
# Copyright (c) 2022 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

import pytest

import libnmstate

from libnmstate.error import NmstateValueError
from libnmstate.schema import Interface
from libnmstate.schema import InterfaceIPv4
from libnmstate.schema import InterfaceIPv6
from libnmstate.schema import InterfaceType
from libnmstate.schema import InterfaceState
from libnmstate.schema import Mptcp

from .testlib import assertlib
from .testlib import cmdlib
from .testlib.env import nm_minor_version


IPV4_ADDRESS1 = "192.0.2.251"
IPV6_ADDRESS1 = "2001:db8:1::1"


@pytest.fixture(scope="module", autouse=True)
def enable_mptcp_in_sysctl():
    ori_value = cmdlib.exec_cmd("sysctl net.mptcp.enabled".split())[1].strip()[
        -1
    ]
    cmdlib.exec_cmd("sysctl -w net.mptcp.enabled=1".split(), check=True)
    yield
    cmdlib.exec_cmd(
        f"sysctl -w net.mptcp.enabled={ori_value}".split(), check=True
    )


@pytest.fixture
def eth1_with_static_ip(eth1_up):
    desired_state = {
        Interface.KEY: [
            {
                Interface.NAME: "eth1",
                Interface.TYPE: InterfaceType.ETHERNET,
                Interface.STATE: InterfaceState.UP,
                Interface.IPV4: {
                    InterfaceIPv4.ENABLED: True,
                    InterfaceIPv4.ADDRESS: [
                        {
                            InterfaceIPv4.ADDRESS_IP: IPV4_ADDRESS1,
                            InterfaceIPv4.ADDRESS_PREFIX_LENGTH: 24,
                        }
                    ],
                },
                Interface.IPV6: {
                    InterfaceIPv6.ENABLED: True,
                    InterfaceIPv6.ADDRESS: [
                        {
                            InterfaceIPv6.ADDRESS_IP: IPV6_ADDRESS1,
                            InterfaceIPv6.ADDRESS_PREFIX_LENGTH: 64,
                        }
                    ],
                },
            }
        ]
    }
    libnmstate.apply(desired_state)
    yield desired_state


@pytest.mark.parametrize(
    "mptcp_flags",
    [
        [Mptcp.FLAG_SIGNAL],
        [Mptcp.FLAG_SUBFLOW],
        [Mptcp.FLAG_BACKUP],
        [Mptcp.FLAG_FULLMESH],
        [Mptcp.FLAG_SIGNAL, Mptcp.FLAG_SUBFLOW],
        [Mptcp.FLAG_SIGNAL, Mptcp.FLAG_BACKUP],
        [Mptcp.FLAG_SUBFLOW, Mptcp.FLAG_BACKUP],
        [Mptcp.FLAG_SUBFLOW, Mptcp.FLAG_FULLMESH],
        [Mptcp.FLAG_SIGNAL, Mptcp.FLAG_SUBFLOW, Mptcp.FLAG_BACKUP],
        [
            Mptcp.FLAG_SUBFLOW,
            Mptcp.FLAG_BACKUP,
            Mptcp.FLAG_FULLMESH,
        ],
        [
            Mptcp.FLAG_FULLMESH,
            Mptcp.FLAG_SUBFLOW,
            Mptcp.FLAG_BACKUP,
        ],
    ],
    ids=[
        "signal",
        "subflow",
        "backup",
        "fullmesh",
        "signal&subflow",
        "signal&backup",
        "subflow@backup",
        "subflow@fullmesh",
        "signal@subflow@backup",
        "subflow@backup@fullmesh",
        "fullmesh@subflow@backup",
    ],
)
@pytest.mark.skipif(
    nm_minor_version() < 40, reason="MPTCP is not supported by NetworkManager"
)
def test_enable_mptcp_flags_and_remove(eth1_with_static_ip, mptcp_flags):
    expected_state = eth1_with_static_ip
    desired_state = {
        Interface.KEY: [
            {
                Interface.NAME: "eth1",
                Interface.TYPE: InterfaceType.ETHERNET,
                Interface.STATE: InterfaceState.UP,
                Interface.MPTCP: {
                    Mptcp.ADDRESS_FLAGS: mptcp_flags,
                },
            }
        ]
    }
    libnmstate.apply(desired_state)

    expected_state[Interface.KEY][0][Interface.MPTCP] = {
        Mptcp.ADDRESS_FLAGS: mptcp_flags
    }
    expected_state[Interface.KEY][0][Interface.IPV4][InterfaceIPv4.ADDRESS][0][
        InterfaceIPv4.MPTCP_FLAGS
    ] = mptcp_flags
    expected_state[Interface.KEY][0][Interface.IPV6][InterfaceIPv6.ADDRESS][0][
        InterfaceIPv6.MPTCP_FLAGS
    ] = mptcp_flags
    assertlib.assert_state_match(expected_state)


@pytest.mark.parametrize(
    "mptcp_flags",
    [
        [Mptcp.FLAG_SIGNAL, Mptcp.FLAG_FULLMESH],
        [Mptcp.FLAG_FULLMESH, Mptcp.FLAG_SIGNAL],
        [Mptcp.FLAG_SIGNAL, Mptcp.FLAG_SUBFLOW, Mptcp.FLAG_FULLMESH],
        [
            Mptcp.FLAG_SIGNAL,
            Mptcp.FLAG_BACKUP,
            Mptcp.FLAG_FULLMESH,
        ],
        [
            Mptcp.FLAG_FULLMESH,
            Mptcp.FLAG_SUBFLOW,
            Mptcp.FLAG_BACKUP,
            Mptcp.FLAG_SIGNAL,
        ],
    ],
    ids=[
        "signal&fullmesh",
        "fullmesh@signal",
        "signal@subflow@fullmesh",
        "signal@backup@fullmesh",
        "fullmesh@subflow@backup@signal",
    ],
)
@pytest.mark.skipif(
    nm_minor_version() < 40, reason="MPTCP is not supported by NetworkManager"
)
def test_invalid_mptcp_flags(eth1_with_static_ip, mptcp_flags):
    with pytest.raises(NmstateValueError):
        libnmstate.apply(
            {
                Interface.KEY: [
                    {
                        Interface.NAME: "eth1",
                        Interface.TYPE: InterfaceType.ETHERNET,
                        Interface.STATE: InterfaceState.UP,
                        Interface.MPTCP: {
                            Mptcp.ADDRESS_FLAGS: mptcp_flags,
                        },
                    }
                ]
            }
        )



#
# Copyright (c) 2021 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from contextlib import contextmanager
import os
import time

import pytest

import libnmstate
from libnmstate.error import NmstateValueError
from libnmstate.schema import Ethtool
from libnmstate.schema import Interface
from libnmstate.schema import InterfaceState
from libnmstate.schema import Veth

from .testlib import assertlib
from .testlib import cmdlib
from .testlib.env import is_fedora
from .testlib.env import nm_major_minor_version

MAX_NETDEVSIM_WAIT_TIME = 5

TEST_NETDEVSIM_NIC = "sim0"


@contextmanager
def netdevsim_interface(ifname):
    try:
        cmdlib.exec_cmd("modprobe netdevsim".split(), check=True)
        with open("/sys/bus/netdevsim/new_device", "w") as fd:
            fd.write("1 1")

        done = False
        for i in range(0, MAX_NETDEVSIM_WAIT_TIME):
            time.sleep(1)
            i += 1
            nics = _get_cur_netdevsim_ifnames()
            if nics:
                _ip_iface_rename(nics[0], ifname)
                done = True
                break
        assert done
        yield
    finally:
        cmdlib.exec_cmd("modprobe -r netdevsim".split())


def _get_cur_netdevsim_ifnames():
    return os.listdir("/sys/devices/netdevsim1/net/")


def _ip_iface_rename(src_name, dst_name):
    cmdlib.exec_cmd(f"ip link set {src_name} down".split(), check=True)
    cmdlib.exec_cmd(
        f"ip link set {src_name} name {dst_name}".split(), check=True
    )


@pytest.mark.skipif(
    nm_major_minor_version() < 1.31
    or os.environ.get("CI") == "true"
    or not is_fedora(),
    reason=(
        "Ethtool pause test need NetworkManager 1.31+ and netdevsim kernel "
        "module"
    ),
)
def test_ethtool_pause_on_netdevsim():
    desire_iface_state = {
        Interface.NAME: TEST_NETDEVSIM_NIC,
        Ethtool.CONFIG_SUBTREE: {
            Ethtool.Pause.CONFIG_SUBTREE: {
                Ethtool.Pause.AUTO_NEGOTIATION: False,
                Ethtool.Pause.RX: True,
                Ethtool.Pause.TX: True,
            }
        },
    }
    with netdevsim_interface(TEST_NETDEVSIM_NIC):
        libnmstate.apply({Interface.KEY: [desire_iface_state]})
        assertlib.assert_state_match({Interface.KEY: [desire_iface_state]})
    assertlib.assert_absent(TEST_NETDEVSIM_NIC)


@pytest.mark.skipif(
    nm_major_minor_version() < 1.31
    or os.environ.get("CI") == "true"
    or not is_fedora(),
    reason=(
        "Ethtool pause test need NetworkManager 1.31+ and netdevsim kernel "
        "module"
    ),
)
def test_ethtool_pause_auto_on_netdevsim():
    desire_iface_state = {
        Interface.NAME: TEST_NETDEVSIM_NIC,
        Ethtool.CONFIG_SUBTREE: {
            Ethtool.Pause.CONFIG_SUBTREE: {
                Ethtool.Pause.AUTO_NEGOTIATION: True,
                Ethtool.Pause.RX: True,
                Ethtool.Pause.TX: True,
            }
        },
    }
    with netdevsim_interface(TEST_NETDEVSIM_NIC):
        libnmstate.apply({Interface.KEY: [desire_iface_state]})
        assertlib.assert_state_match({Interface.KEY: [desire_iface_state]})
    assertlib.assert_absent(TEST_NETDEVSIM_NIC)


@pytest.mark.skipif(
    os.environ.get("CI") == "true",
    reason=("CI does not have ethtool kernel option enabled"),
)
def test_ethtool_feature_using_ethtool_cli_alias_rx_checksumming(eth1_up):
    desire_iface_state = {
        Interface.NAME: "eth1",
        Ethtool.CONFIG_SUBTREE: {
            Ethtool.Feature.CONFIG_SUBTREE: {"rx-checksumming": False}
        },
    }
    libnmstate.apply({Interface.KEY: [desire_iface_state]})

    desire_feature = desire_iface_state[Ethtool.CONFIG_SUBTREE][
        Ethtool.Feature.CONFIG_SUBTREE
    ]
    desire_feature.pop("rx-checksumming")
    desire_feature["rx-checksum"] = False

    assertlib.assert_state_match({Interface.KEY: [desire_iface_state]})


def test_ethtool_invalid_feature(eth1_up):
    desire_iface_state = {
        Interface.NAME: "eth1",
        Ethtool.CONFIG_SUBTREE: {
            Ethtool.Feature.CONFIG_SUBTREE: {"no_exist_feature": False}
        },
    }
    with pytest.raises(NmstateValueError):
        libnmstate.apply({Interface.KEY: [desire_iface_state]})


@pytest.mark.skipif(
    not os.environ.get("TEST_REAL_NIC"),
    reason="Need to define TEST_REAL_NIC for ethtool ring test",
)
def test_ethtool_ring_set_rx():
    desire_iface_state = {
        Interface.NAME: os.environ.get("TEST_REAL_NIC"),
        Ethtool.CONFIG_SUBTREE: {
            Ethtool.Ring.CONFIG_SUBTREE: {Ethtool.Ring.RX: 256}
        },
    }
    libnmstate.apply({Interface.KEY: [desire_iface_state]})

    assertlib.assert_state_match({Interface.KEY: [desire_iface_state]})


@pytest.mark.skipif(
    nm_major_minor_version() < 1.31
    or os.environ.get("CI") == "true"
    or not is_fedora(),
    reason=(
        "Ethtool pause test need NetworkManager 1.31+ and netdevsim kernel "
        "module in Fedora 34+"
    ),
)
def test_ethtool_coalesce_on_netdevsim():
    desire_iface_state = {
        Interface.NAME: TEST_NETDEVSIM_NIC,
        Ethtool.CONFIG_SUBTREE: {
            Ethtool.Coalesce.CONFIG_SUBTREE: {
                Ethtool.Coalesce.TX_USECS: 100,
            }
        },
    }
    with netdevsim_interface(TEST_NETDEVSIM_NIC):
        libnmstate.apply({Interface.KEY: [desire_iface_state]})
        assertlib.assert_state_match({Interface.KEY: [desire_iface_state]})
    assertlib.assert_absent(TEST_NETDEVSIM_NIC)


@pytest.fixture
def veth1_with_ethtool_feature_highdma_false():
    interface_name = "veth1"
    peer_interface_name = f"{interface_name}.ep"
    iface_state = {
        Interface.NAME: interface_name,
        Interface.TYPE: Veth.TYPE,
        Interface.STATE: InterfaceState.UP,
        Veth.CONFIG_SUBTREE: {Veth.PEER: peer_interface_name},
        Ethtool.CONFIG_SUBTREE: {
            Ethtool.Feature.CONFIG_SUBTREE: {
                "highdma": False,
            }
        },
    }
    libnmstate.apply({Interface.KEY: [iface_state]})
    yield iface_state
    libnmstate.apply(
        {
            Interface.KEY: [
                {
                    Interface.NAME: interface_name,
                    Interface.STATE: InterfaceState.ABSENT,
                },
                {
                    Interface.NAME: peer_interface_name,
                    Interface.STATE: InterfaceState.ABSENT,
                },
            ],
        },
        verify_change=False,
    )


@pytest.fixture
def veth1_with_ethtool_feature_highdma_true():
    interface_name = "veth1"
    peer_interface_name = f"{interface_name}.ep"
    iface_state = {
        Interface.NAME: interface_name,
        Interface.TYPE: Veth.TYPE,
        Interface.STATE: InterfaceState.UP,
        Veth.CONFIG_SUBTREE: {Veth.PEER: peer_interface_name},
        Ethtool.CONFIG_SUBTREE: {
            Ethtool.Feature.CONFIG_SUBTREE: {
                "highdma": True,
            }
        },
    }
    libnmstate.apply({Interface.KEY: [iface_state]})
    yield iface_state
    libnmstate.apply(
        {
            Interface.KEY: [
                {
                    Interface.NAME: interface_name,
                    Interface.STATE: InterfaceState.ABSENT,
                },
                {
                    Interface.NAME: peer_interface_name,
                    Interface.STATE: InterfaceState.ABSENT,
                },
            ],
        },
        verify_change=False,
    )


@pytest.mark.skipif(
    os.environ.get("CI") == "true",
    reason=("CI environment does not support ethtool via netlink yet"),
)
def test_ethtool_preserve_existing_ethtool_feature_setting(
    veth1_with_ethtool_feature_highdma_false,
):
    iface_state = veth1_with_ethtool_feature_highdma_false

    libnmstate.apply(
        {
            Interface.KEY: [
                {
                    Interface.NAME: iface_state[Interface.NAME],
                    Interface.MTU: 1400,
                }
            ]
        }
    )
    iface_state[Interface.MTU] = 1400
    assertlib.assert_state_match({Interface.KEY: [iface_state]})



#
# Copyright (c) 2022 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

import time

import pytest
import yaml

import libnmstate
from libnmstate.schema import Interface
from libnmstate.schema import InterfaceType
from libnmstate.schema import OVSBridge
from libnmstate.schema import RouteRule

from ..testlib.env import is_k8s
from ..testlib import iprule
from ..testlib.statelib import show_only
from ..testlib.genconf import gen_conf_apply


NM_CONFIG_FOLDER = "/etc/NetworkManager/system-connections"
MAX_RETRY_COUNT = 20


@pytest.fixture
def cleanup_ovs_same_name():
    yield
    libnmstate.apply(
        load_yaml(
            """
interfaces:
- name: br0
  type: ovs-bridge
  state: absent
"""
        ),
        verify_change=False,
    )


@pytest.mark.tier1
@pytest.mark.skipif(is_k8s(), reason="K8S does not support genconf")
def test_gen_conf_ovs_same_name(eth1_up, cleanup_ovs_same_name):
    desired_state = load_yaml(
        """
interfaces:
- name: eth1
  type: ethernet
  state: up
- name: br0
  type: ovs-bridge
  state: up
  bridge:
    port:
    - name: eth1
    - name: br0
- name: br0
  type: ovs-interface
  state: up
"""
    )

    with gen_conf_apply(desired_state):
        retry_verify_ovs_ports("br0", sorted(["eth1", "br0"]))


@pytest.mark.tier1
def test_gen_conf_routes_rules():
    desired_state = load_yaml(
        """
interfaces:
- name: eth1
  type: ethernet
  state: up
  ipv4:
    address:
      - ip: 192.0.2.251
        prefix-length: 24
    dhcp: false
    enabled: true
routes:
  config:
    - destination: 198.51.100.0/24
      metric: 150
      next-hop-address: 192.0.2.1
      next-hop-interface: eth1
      table-id: 254
route-rules:
  config:
    - ip-to: 192.0.2.0/24
      ip-from: 192.168.2.0/24
      priority: 1
      route-table: 254
"""
    )
    with gen_conf_apply(desired_state):
        rule = desired_state[RouteRule.KEY][RouteRule.CONFIG][0]
        iprule.ip_rule_exist_in_os(
            rule.get(RouteRule.IP_FROM),
            rule.get(RouteRule.IP_TO),
            rule.get(RouteRule.PRIORITY),
            rule.get(RouteRule.ROUTE_TABLE),
            rule.get(RouteRule.FWMARK),
            rule.get(RouteRule.FWMASK),
            rule.get(RouteRule.FAMILY),
        )


def load_yaml(content):
    return yaml.load(content, Loader=yaml.SafeLoader)


# the assert_state_match does not works well on OVS same name
# manual checking
def retry_verify_ovs_ports(bridge_name, port_names):
    retry_count = 0
    while retry_count < MAX_RETRY_COUNT:
        try:
            verify_ovs_ports(bridge_name, port_names)
            break
        except AssertionError:
            retry_count += 1
            time.sleep(1)

    verify_ovs_ports(bridge_name, port_names)


def verify_ovs_ports(bridge_name, port_names):
    cur_iface = None
    for iface in show_only((bridge_name,))[Interface.KEY]:
        if iface[Interface.TYPE] == InterfaceType.OVS_BRIDGE:
            cur_iface = iface
            break
    assert cur_iface

    cur_ports = [
        p[OVSBridge.Port.NAME]
        for p in cur_iface[OVSBridge.CONFIG_SUBTREE][OVSBridge.PORT_SUBTREE]
    ]
    cur_ports.sort()
    assert cur_ports == port_names



# SPDX-License-Identifier: LGPL-2.1-or-later

import json

import libnmstate
from libnmstate.schema import Interface

from .cmdlib import exec_cmd


def get_sriov_vf_names(pf_name):
    output = exec_cmd(f"ip -j -d link show {pf_name}".split())[1]
    link_info = json.loads(output)[0]
    macs = [
        vf_info["address"].upper()
        for vf_info in link_info.get("vfinfo_list", [])
    ]
    return [
        iface[Interface.NAME]
        for iface in libnmstate.show()[Interface.KEY]
        if iface[Interface.MAC] in macs
    ]



# SPDX-License-Identifier: LGPL-2.1-or-later

import libnmstate
import pytest
import yaml

from ..testlib.env import nm_minor_version
from ..testlib.cmdlib import exec_cmd


@pytest.fixture
def cleanup_loopback():
    yield
    libnmstate.apply(
        yaml.load(
            """---
            interfaces:
            - name: lo
              type: loopback
              state: absent
            """,
            Loader=yaml.SafeLoader,
        )
    )


@pytest.mark.skipif(
    nm_minor_version() < 41,
    reason=("Loopback is only supported by NetworkManager 1.41+"),
)
def test_store_route_table_local_rule_to_loopback(cleanup_loopback):
    libnmstate.apply(
        yaml.load(
            """---
            route-rules:
              config:
                - route-table: 255
                  priority: 12345
                  family: ipv4
                - route-table: 255
                  priority: 12346
                  family: ipv6
            """,
            Loader=yaml.SafeLoader,
        )
    )

    assert (
        exec_cmd("nmcli -g ipv4.routing-rules c show lo".split())[1].strip()
        == "priority 12345 from 0.0.0.0/0 table 255"
    )

    assert (
        exec_cmd("nmcli -g ipv6.routing-rules c show lo".split())[1].strip()
        == r"priority 12346 from \:\:/0 table 255"
    )



#
# Copyright (c) 2018-2021 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
""" Test the nmstate example files """
import os

import pytest

from .testlib import assertlib
from .testlib import iprule
from .testlib.examplelib import example_state
from .testlib.examplelib import find_examples_dir
from .testlib.examplelib import load_example

import libnmstate
from libnmstate import netinfo
from libnmstate.error import NmstateNotSupportedError
from libnmstate.error import NmstateVerificationError
from libnmstate.schema import DNS
from libnmstate.schema import HostNameState
from libnmstate.schema import RouteRule

from .testlib.env import is_k8s
from .testlib.env import nm_major_minor_version


@pytest.mark.tier1
def test_add_down_remove_vlan(eth1_up):
    """
    Test adding, downing and removing a vlan
    """

    vlan_ifname = "eth1.101"
    with example_state(
        "vlan101_eth1_up.yml", cleanup="vlan101_eth1_absent.yml"
    ) as desired_state:
        assertlib.assert_state(desired_state)
        with example_state("vlan101_eth1_down.yml") as desired_state:
            assertlib.assert_absent(vlan_ifname)

    assertlib.assert_absent(vlan_ifname)


@pytest.mark.tier1
@pytest.mark.xfail(
    is_k8s(),
    reason=(
        "Requires adjusts for k8s. Ref:"
        "https://github.com/nmstate/nmstate/issues/1579"
    ),
    raises=NmstateVerificationError,
    strict=False,
)
def test_add_remove_ovs_bridge(eth1_up):
    with example_state(
        "ovsbridge_create.yml", cleanup="ovsbridge_delete.yml"
    ) as desired_state:
        assertlib.assert_state(desired_state)

    assertlib.assert_absent("ovs-br0")


@pytest.mark.tier1
def test_add_remove_ovs_bridge_bond(eth1_up, eth2_up):
    with example_state(
        "ovsbridge_bond_create.yml", cleanup="ovsbridge_delete.yml"
    ) as desired_state:
        assertlib.assert_state_match(desired_state)

    assertlib.assert_absent("ovs-br0")
    assertlib.assert_absent("ovs-bond1")


@pytest.mark.tier1
def test_add_remove_ovs_bridge_vlan(eth1_up, eth2_up):
    with example_state(
        "ovsbridge_vlan_port.yml", cleanup="ovsbridge_delete.yml"
    ) as desired_state:
        assertlib.assert_state_match(desired_state)

    assertlib.assert_absent("ovs-br0")


@pytest.mark.tier1
def test_add_remove_linux_bridge(eth1_up):
    with example_state(
        "linuxbrige_eth1_up.yml", cleanup="linuxbrige_eth1_absent.yml"
    ) as desired_state:
        assertlib.assert_state(desired_state)

    assertlib.assert_absent("linux-br0")


@pytest.mark.tier1
def test_bond_linuxbridge_vlan(eth1_up, eth2_up):
    with example_state(
        "bond_linuxbridge_vlan_up.yml",
        cleanup="bond_linuxbridge_vlan_absent.yml",
    ) as desired_state:
        assertlib.assert_state_match(desired_state)

    assertlib.assert_absent("bond0")
    assertlib.assert_absent("br0")
    assertlib.assert_absent("br29")
    assertlib.assert_absent("vlan29")


@pytest.mark.tier1
def test_dns_edit(eth1_up):
    with example_state(
        "dns_edit_eth1.yml", cleanup="dns_remove.yml"
    ) as desired_state:
        assertlib.assert_state_match(desired_state)

    current_state = netinfo.show()
    assert current_state.get(DNS.KEY, {}).get(DNS.CONFIG, {}) == {}


@pytest.mark.tier1
def test_add_remove_routes(eth1_up):
    """
    Test adding a strict route and removing all routes next hop to eth1.
    """
    with example_state(
        "eth1_add_route.yml", cleanup="eth1_del_all_routes.yml"
    ) as desired_state:
        assertlib.assert_state(desired_state)

    assertlib.assert_no_config_route_to_iface("eth1")


@pytest.mark.tier1
def test_add_remove_route_rule(eth1_up):
    """
    Test adding a route rule and removing all route rules next hop to eth1.
    """
    with example_state(
        "eth1_add_route_rule.yml", cleanup="eth1_del_all_route_rules.yml"
    ) as desired_state:
        rule = desired_state[RouteRule.KEY][RouteRule.CONFIG][0]
        iprule.ip_rule_exist_in_os(
            rule.get(RouteRule.IP_FROM),
            rule.get(RouteRule.IP_TO),
            rule.get(RouteRule.PRIORITY),
            rule.get(RouteRule.ROUTE_TABLE),
            rule.get(RouteRule.FWMARK),
            rule.get(RouteRule.FWMASK),
            rule.get(RouteRule.FAMILY),
        )


@pytest.mark.skipif(
    os.environ.get("CI") == "true",
    reason="SR-IOV device required for this test case",
)
@pytest.mark.xfail(
    raises=NmstateNotSupportedError,
    reason="The device does not support SR-IOV.",
)
def test_set_ethernet_sriov(eth1_up):
    with example_state("eth1_with_sriov.yml") as desired_state:
        assertlib.assert_state_match(desired_state)


def test_port_vlan(eth1_up):
    with example_state(
        "linuxbrige_eth1_up_port_vlan.yml",
        cleanup="linuxbrige_eth1_absent.yml",
    ) as desired_state:
        assertlib.assert_state(desired_state)

    assertlib.assert_absent("linux-br0")


def test_add_ovs_patch_and_remove():
    with example_state(
        "ovsbridge_patch_create.yml", cleanup="ovsbridge_patch_delete.yml"
    ) as desired_state:
        assertlib.assert_state(desired_state)

    assertlib.assert_absent("patch0")


def test_add_remove_vrf(eth1_up):
    with example_state(
        "vrf0_with_eth1.yml", cleanup="vrf0_absent.yml"
    ) as desired_state:
        assertlib.assert_state_match(desired_state)

    assertlib.assert_absent("vrf0")


@pytest.mark.skipif(
    not os.environ.get("TEST_REAL_NIC"),
    reason="Need to define TEST_REAL_NIC for infiniband test",
)
def test_add_ib_pkey_nic_and_remove():
    test_nic = os.environ["TEST_REAL_NIC"]
    with example_state(
        "infiniband_pkey_ipoib_create.yml",
        cleanup="infiniband_pkey_ipoib_delete.yml",
        substitute=("mlx5_ib0", test_nic),
    ) as desired_state:
        assertlib.assert_state(desired_state)

    assertlib.assert_absent(f"{test_nic}.80ff")


def test_add_mac_vlan_and_remove():
    with example_state(
        "mac_vlan_create.yml", cleanup="mac_vlan_absent.yml"
    ) as desired_state:
        assertlib.assert_state(desired_state)

    assertlib.assert_absent("macvlan0")


def test_add_mac_vtap_and_remove():
    with example_state(
        "mac_vtap_create.yml", cleanup="mac_vtap_absent.yml"
    ) as desired_state:
        assertlib.assert_state(desired_state)

    assertlib.assert_absent("macvtap0")


@pytest.mark.skipif(
    nm_major_minor_version() <= 1.28,
    reason="Modifying veth is not supported on lower NetworkManager versions.",
)
def test_add_veth_and_remove():
    with example_state(
        "veth1_up.yml", cleanup="veth1_absent.yml"
    ) as desired_state:
        assertlib.assert_state(desired_state)

    assertlib.assert_absent("veth1")
    assertlib.assert_absent("veth1peer")


@pytest.mark.skipif(
    nm_major_minor_version() <= 1.30,
    reason="Generating config is not supported on NetworkManager 1.30-",
)
@pytest.mark.tier1
def test_gen_conf_for_examples():
    example_dir = find_examples_dir()
    with os.scandir(example_dir) as example_dir_fd:
        for example_file in example_dir_fd:
            if example_file.name.endswith(".yml"):
                first_result = libnmstate.generate_configurations(
                    load_example(example_file.name)
                )
                second_result = libnmstate.generate_configurations(
                    load_example(example_file.name)
                )
                assert first_result == second_result


@pytest.mark.skipif(
    os.environ.get("CI") == "true",
    reason="NM cannot change hostname in container",
)
def test_static_hostname_for_examples():
    with example_state(
        "static_hostname.yml", cleanup="dynamic_hostname.yml"
    ) as desired_state:
        cur_hostname = libnmstate.show()[HostNameState.KEY]
        assert cur_hostname == desired_state[HostNameState.KEY]



# SPDX-License-Identifier: LGPL-2.1-or-later

import pytest
import yaml

import libnmstate

from libnmstate.error import NmstateValueError
from libnmstate.schema import Interface
from libnmstate.schema import InterfaceIPv4
from libnmstate.schema import InterfaceIPv6
from libnmstate.schema import InterfaceType
from libnmstate.schema import InterfaceState
from libnmstate.schema import VRF

from .testlib import assertlib
from .testlib import cmdlib


TEST_VRF0 = "test-vrf0"
TEST_VRF1 = "test-vrf1"
TEST_VRF_PORT0 = "eth1"
TEST_VRF_PORT1 = "eth2"
TEST_VRF_VETH0 = "veth0"
TEST_VRF_VETH1 = "veth1"
TEST_ROUTE_TABLE_ID0 = 100
TEST_ROUTE_TABLE_ID1 = 101
IPV4_ADDRESS1 = "192.0.2.251"
IPV6_ADDRESS1 = "2001:db8:1::1"
TEST_MAC_ADDRESS = "00:00:5E:00:53:01"
TEST_BOND0 = "test-bond0"
TEST_BOND0_VLAN = "test-bond0.100"


@pytest.fixture
def vrf0_with_port0(port1_up):
    vrf_iface_info = {
        Interface.NAME: TEST_VRF0,
        Interface.TYPE: InterfaceType.VRF,
        VRF.CONFIG_SUBTREE: {
            VRF.PORT_SUBTREE: [TEST_VRF_PORT0],
            VRF.ROUTE_TABLE_ID: TEST_ROUTE_TABLE_ID0,
        },
    }
    libnmstate.apply({Interface.KEY: [vrf_iface_info]})
    yield vrf_iface_info
    libnmstate.apply(
        {
            Interface.KEY: [
                {
                    Interface.NAME: TEST_VRF0,
                    Interface.STATE: InterfaceState.ABSENT,
                }
            ]
        }
    )
    assertlib.assert_absent(TEST_VRF0)


@pytest.fixture
def vrf1_with_port1(port1_up):
    vrf_iface_info = {
        Interface.NAME: TEST_VRF1,
        Interface.TYPE: InterfaceType.VRF,
        VRF.CONFIG_SUBTREE: {
            VRF.PORT_SUBTREE: [TEST_VRF_PORT1],
            VRF.ROUTE_TABLE_ID: TEST_ROUTE_TABLE_ID1,
        },
    }
    libnmstate.apply({Interface.KEY: [vrf_iface_info]})
    yield vrf_iface_info
    libnmstate.apply(
        {
            Interface.KEY: [
                {
                    Interface.NAME: TEST_VRF1,
                    Interface.STATE: InterfaceState.ABSENT,
                }
            ]
        }
    )
    assertlib.assert_absent(TEST_VRF1)


@pytest.fixture
def vrf1_with_eth1_and_eth2(eth1_up, eth2_up):
    vrf_iface_info = {
        Interface.NAME: TEST_VRF1,
        Interface.TYPE: InterfaceType.VRF,
        VRF.CONFIG_SUBTREE: {
            VRF.PORT_SUBTREE: ["eth1", "eth2"],
            VRF.ROUTE_TABLE_ID: TEST_ROUTE_TABLE_ID1,
        },
    }
    libnmstate.apply({Interface.KEY: [vrf_iface_info]})
    yield vrf_iface_info
    libnmstate.apply(
        {
            Interface.KEY: [
                {
                    Interface.NAME: TEST_VRF1,
                    Interface.STATE: InterfaceState.ABSENT,
                }
            ]
        }
    )
    assertlib.assert_absent(TEST_VRF1)


@pytest.fixture
def unmanaged_port_up():
    cmdlib.exec_cmd(
        f"ip link add {TEST_VRF_VETH0} type veth peer {TEST_VRF_VETH1}".split()
    )
    cmdlib.exec_cmd(f"ip link set {TEST_VRF_VETH0} up".split())
    cmdlib.exec_cmd(f"ip link set {TEST_VRF_VETH1} up".split())
    try:
        yield TEST_VRF_VETH0
    finally:
        cmdlib.exec_cmd(f"ip link del {TEST_VRF_VETH0}".split())


@pytest.fixture
def vrf1_with_unmanaged_port(unmanaged_port_up):
    vrf_iface_info = {
        Interface.NAME: TEST_VRF1,
        Interface.TYPE: InterfaceType.VRF,
        VRF.CONFIG_SUBTREE: {
            VRF.PORT_SUBTREE: [unmanaged_port_up],
            VRF.ROUTE_TABLE_ID: TEST_ROUTE_TABLE_ID1,
        },
    }
    veth_iface_info = {
        Interface.NAME: unmanaged_port_up,
        Interface.TYPE: InterfaceType.ETHERNET,
        Interface.STATE: InterfaceState.UP,
    }
    libnmstate.apply({Interface.KEY: [vrf_iface_info, veth_iface_info]})
    try:
        yield vrf_iface_info
    finally:
        libnmstate.apply(
            {
                Interface.KEY: [
                    {
                        Interface.NAME: TEST_VRF1,
                        Interface.STATE: InterfaceState.ABSENT,
                    }
                ]
            }
        )


@pytest.fixture
def vrf_over_bond_vlan(eth1_up, eth2_up):
    desired = yaml.load(
        """---
        interfaces:
        - name: test-bond0.100
          type: vlan
          vlan:
            base-iface: test-bond0
            id: 100
        - name: test-bond0
          type: bond
          link-aggregation:
            mode: balance-rr
        - name: test-vrf0
          type: vrf
          state: up
          vrf:
            port:
            - test-bond0
            - test-bond0.100
            route-table-id: 100
        """,
        Loader=yaml.SafeLoader,
    )
    libnmstate.apply(desired)
    yield desired
    libnmstate.apply(
        yaml.load(
            """---
            interfaces:
            - name: test-bond0.100
              type: vlan
              state: absent
            - name: test-bond0
              type: bond
              state: absent
            - name: test-vrf0
              type: vrf
              state: absent
            """,
            Loader=yaml.SafeLoader,
        )
    )


class TestVrf:
    def test_create_and_remove(self, vrf0_with_port0):
        pass

    def test_sort_ports(self, vrf1_with_eth1_and_eth2):
        iface_info = vrf1_with_eth1_and_eth2
        iface_info[VRF.CONFIG_SUBTREE][VRF.PORT_SUBTREE].reverse()
        desired_state = {Interface.KEY: [iface_info]}
        libnmstate.apply(desired_state)
        iface_info[VRF.CONFIG_SUBTREE][VRF.PORT_SUBTREE].sort()
        assertlib.assert_state_match({Interface.KEY: [iface_info]})

    def test_change_route_table_id(self, vrf0_with_port0):
        iface_info = vrf0_with_port0
        iface_info[VRF.CONFIG_SUBTREE][VRF.ROUTE_TABLE_ID] += 1
        desired_state = {Interface.KEY: [iface_info]}
        libnmstate.apply(desired_state)
        assertlib.assert_state_match(desired_state)

    def test_create_with_empty_ports(self):
        desired_state = {
            Interface.KEY: [
                {
                    Interface.NAME: TEST_VRF0,
                    Interface.TYPE: InterfaceType.VRF,
                    VRF.CONFIG_SUBTREE: {
                        VRF.PORT_SUBTREE: [],
                        VRF.ROUTE_TABLE_ID: TEST_ROUTE_TABLE_ID0,
                    },
                }
            ]
        }
        libnmstate.apply(desired_state)
        assertlib.assert_state_match(desired_state)

    def test_add_and_remove_port(self, vrf0_with_port0, port1_up):
        iface_info = vrf0_with_port0
        iface_info[VRF.CONFIG_SUBTREE][VRF.PORT_SUBTREE].append(TEST_VRF_PORT1)
        desired_state = {Interface.KEY: [iface_info]}
        libnmstate.apply(desired_state)
        assertlib.assert_state_match(desired_state)

        iface_info[VRF.CONFIG_SUBTREE][VRF.PORT_SUBTREE].remove(TEST_VRF_PORT1)
        desired_state = {Interface.KEY: [iface_info]}
        libnmstate.apply(desired_state)
        assertlib.assert_state_match(desired_state)

    def test_remove_port(self, vrf0_with_port0):
        iface_info = vrf0_with_port0
        iface_info[VRF.CONFIG_SUBTREE][VRF.PORT_SUBTREE] = []
        desired_state = {Interface.KEY: [iface_info]}
        libnmstate.apply(desired_state)
        assertlib.assert_state_match(desired_state)

    def test_remove_all_ports(self, vrf0_with_port0):
        iface_info = vrf0_with_port0
        iface_info[VRF.CONFIG_SUBTREE][VRF.PORT_SUBTREE] = []
        desired_state = {Interface.KEY: [iface_info]}
        libnmstate.apply(desired_state)
        assertlib.assert_state_match(desired_state)

    def test_moving_port_from_other_vrf(
        self, vrf0_with_port0, vrf1_with_port1
    ):
        vrf0_iface = vrf0_with_port0
        vrf1_iface = vrf1_with_port1
        vrf0_iface[VRF.CONFIG_SUBTREE][VRF.PORT_SUBTREE] = []
        vrf1_iface[VRF.CONFIG_SUBTREE][VRF.PORT_SUBTREE] = [
            TEST_VRF_PORT0,
            TEST_VRF_PORT1,
        ]

        desired_state = {Interface.KEY: [vrf0_iface, vrf1_iface]}
        libnmstate.apply(desired_state)
        assertlib.assert_state_match(desired_state)

    def test_swaping_port(self, vrf0_with_port0, vrf1_with_port1):
        vrf0_iface = vrf0_with_port0
        vrf1_iface = vrf1_with_port1
        vrf0_iface[VRF.CONFIG_SUBTREE][VRF.PORT_SUBTREE] = [TEST_VRF_PORT1]
        vrf1_iface[VRF.CONFIG_SUBTREE][VRF.PORT_SUBTREE] = [TEST_VRF_PORT0]
        desired_state = {Interface.KEY: [vrf0_iface, vrf1_iface]}
        libnmstate.apply(desired_state)
        assertlib.assert_state_match(desired_state)

    def test_port_holding_ip(self, vrf0_with_port0):
        desired_state = {
            Interface.KEY: [
                {
                    Interface.NAME: TEST_VRF_PORT0,
                    Interface.IPV4: {
                        InterfaceIPv4.ENABLED: True,
                        InterfaceIPv4.DHCP: False,
                        InterfaceIPv4.ADDRESS: [
                            {
                                InterfaceIPv4.ADDRESS_IP: IPV4_ADDRESS1,
                                InterfaceIPv4.ADDRESS_PREFIX_LENGTH: 24,
                            }
                        ],
                    },
                    Interface.IPV6: {
                        InterfaceIPv6.ENABLED: True,
                        InterfaceIPv6.DHCP: False,
                        InterfaceIPv6.AUTOCONF: False,
                        InterfaceIPv6.ADDRESS: [
                            {
                                InterfaceIPv6.ADDRESS_IP: IPV6_ADDRESS1,
                                InterfaceIPv6.ADDRESS_PREFIX_LENGTH: 64,
                            }
                        ],
                    },
                }
            ]
        }
        libnmstate.apply(desired_state)
        assertlib.assert_state_match(desired_state)

    def test_takes_over_unmanaged_vrf(self, vrf1_with_unmanaged_port):
        pass

    def test_vrf_ignore_mac_address(self, vrf0_with_port0):
        libnmstate.apply(
            {
                Interface.KEY: [
                    {
                        Interface.NAME: TEST_VRF0,
                        Interface.MAC: TEST_MAC_ADDRESS,
                    }
                ]
            }
        )

    def test_vrf_ignore_accept_all_mac_addresses_false(self, vrf0_with_port0):
        libnmstate.apply(
            {
                Interface.KEY: [
                    {
                        Interface.NAME: TEST_VRF0,
                        Interface.ACCEPT_ALL_MAC_ADDRESSES: False,
                    }
                ]
            }
        )

    def test_change_vrf_without_table_id(self, vrf0_with_port0):
        libnmstate.apply(
            {
                Interface.KEY: [
                    {
                        Interface.NAME: TEST_VRF0,
                        VRF.CONFIG_SUBTREE: {
                            VRF.PORT_SUBTREE: [TEST_VRF_PORT0],
                        },
                    }
                ]
            }
        )

    def test_new_vrf_without_table_id(self):
        with pytest.raises(NmstateValueError):
            libnmstate.apply(
                {
                    Interface.KEY: [
                        {
                            Interface.NAME: TEST_VRF0,
                            VRF.CONFIG_SUBTREE: {
                                VRF.PORT_SUBTREE: [TEST_VRF_PORT0],
                            },
                        }
                    ]
                }
            )

    def test_new_vrf_over_new_bond_vlan(self, vrf_over_bond_vlan):
        assertlib.assert_state_match(vrf_over_bond_vlan)

    def test_vrf_over_bond_vlan_got_auto_remove_by_parent(
        self, vrf_over_bond_vlan
    ):
        libnmstate.apply(
            {
                Interface.KEY: [
                    {
                        Interface.NAME: TEST_VRF0,
                        Interface.STATE: InterfaceState.ABSENT,
                    },
                    {
                        Interface.NAME: TEST_BOND0,
                        Interface.STATE: InterfaceState.ABSENT,
                    },
                ]
            }
        )
        assertlib.assert_absent(TEST_VRF0)
        assertlib.assert_absent(TEST_BOND0)
        assertlib.assert_absent(TEST_BOND0_VLAN)



# SPDX-License-Identifier: LGPL-2.1-or-later

from contextlib import contextmanager
import os
import pytest
import yaml

import libnmstate
from libnmstate.prettystate import PrettyState
from libnmstate.schema import Interface
from libnmstate.schema import InterfaceIP
from libnmstate.schema import InterfaceIPv4
from libnmstate.schema import InterfaceState
from libnmstate.schema import InterfaceType
from libnmstate.schema import LinuxBridge
from libnmstate.schema import MacVlan
from libnmstate.schema import MacVtap
from libnmstate.schema import OVSBridge
from libnmstate.schema import OVSInterface
from libnmstate.schema import OvsDB
from libnmstate.schema import RouteRule
from libnmstate.schema import VLAN
from libnmstate.schema import VXLAN
from libnmstate.schema import Veth
from libnmstate.error import NmstateDependencyError
from libnmstate.error import NmstateNotSupportedError
from libnmstate.error import NmstateValueError

from .testlib import assertlib
from .testlib import cmdlib
from .testlib import iprule
from .testlib import statelib
from .testlib.bondlib import bond_interface
from .testlib.bridgelib import linux_bridge
from .testlib.env import is_k8s
from .testlib.env import nm_major_minor_version
from .testlib.env import nm_minor_version
from .testlib.genconf import gen_conf_apply
from .testlib.nmplugin import disable_nm_plugin
from .testlib.ovslib import Bridge
from .testlib.retry import retry_till_true_or_timeout
from .testlib.servicelib import disable_service
from .testlib.statelib import state_match
from .testlib.vlan import vlan_interface


BOND1 = "bond1"
BRIDGE0 = "br0"
BRIDGE1 = "br1"
PORT1 = "ovs1"
PORT2 = "ovs2"
PATCH0 = "patch0"
PATCH1 = "patch1"
VLAN_IFNAME = "eth101"
VETH1 = "test-veth1"
VETH1_PEER = "test-veth1-ep"

MAC1 = "02:FF:FF:FF:FF:01"

ETH1 = "eth1"
ETH2 = "eth2"

EMPTY_MAP = "{}"

OVS_BOND_YAML_STATE = f"""
    port:
    - name: {BOND1}
      link-aggregation:
        mode: active-backup
        port:
        - name: {ETH1}
        - name: {ETH2}
"""

RC_SUCCESS = 0
TEST_EXTERNAL_IDS_KEY = "ovn-bridge-mappings"
TEST_EXTERNAL_IDS_VALUE = "provider:br-provider"
TEST_OTHER_CONFIG_KEY = "stats-update-interval"
TEST_OTHER_CONFIG_VALUE = "1000"
RETRY_TIMEOUT = 15


def _test_pci_path():
    return os.environ.get("TEST_PCI_PATH")


@pytest.fixture
def bridge_with_ports(port0_up):
    system_port0_name = port0_up[Interface.KEY][0][Interface.NAME]

    bridge = Bridge(BRIDGE1)
    bridge.add_system_port(system_port0_name)
    bridge.add_internal_port(PORT1, ipv4_state={InterfaceIPv4.ENABLED: False})
    with bridge.create():
        yield bridge


@pytest.mark.tier1
def test_create_and_remove_ovs_bridge_with_min_desired_state():
    with Bridge(BRIDGE1).create():
        # assert_state_match does not works well when ovs bridge and ovs
        # interface are using the same name.
        cur_state = statelib.show_only((BRIDGE1,))
        assert len(cur_state[Interface.KEY]) == 2

    assertlib.assert_absent(BRIDGE1)


@pytest.mark.tier1
def test_create_and_save_ovs_bridge_then_remove_and_apply_again():
    desired_state = {}
    with Bridge(BRIDGE1).create():
        desired_state = statelib.show_only((BRIDGE1,))

    assertlib.assert_absent(BRIDGE1)

    libnmstate.apply(desired_state)
    desired_state[Interface.KEY][0][Interface.STATE] = InterfaceState.ABSENT
    desired_state[Interface.KEY][1][Interface.STATE] = InterfaceState.ABSENT

    libnmstate.apply(desired_state)
    assertlib.assert_absent(BRIDGE1)


def test_create_and_remove_ovs_bridge_options_specified():
    bridge = Bridge(BRIDGE1)
    bridge.set_options(
        {
            OVSBridge.Options.FAIL_MODE: "",
            OVSBridge.Options.MCAST_SNOOPING_ENABLED: False,
            OVSBridge.Options.RSTP: False,
            OVSBridge.Options.STP: True,
        }
    )

    bridge.add_internal_port(PORT1, ipv4_state={InterfaceIPv4.ENABLED: False})
    with bridge.create() as state:
        assertlib.assert_state_match(state)

    assertlib.assert_absent(BRIDGE1)


@pytest.mark.tier1
def test_create_and_remove_ovs_bridge_with_a_system_port(port0_up):
    bridge = Bridge(BRIDGE1)
    port0_name = port0_up[Interface.KEY][0][Interface.NAME]
    bridge.add_system_port(port0_name)

    with bridge.create() as state:
        assertlib.assert_state_match(state)

    assertlib.assert_absent(BRIDGE1)

    state = statelib.show_only((port0_name,))
    assert state
    assert state[Interface.KEY][0][Interface.STATE] == InterfaceState.UP


@pytest.mark.tier1
def test_create_and_remove_ovs_bridge_with_internal_port_static_ip_and_mac():
    bridge = Bridge(BRIDGE1)
    bridge.add_internal_port(
        PORT1,
        mac=MAC1,
        ipv4_state={
            InterfaceIPv4.ENABLED: True,
            InterfaceIPv4.ADDRESS: [
                {
                    InterfaceIPv4.ADDRESS_IP: "192.0.2.1",
                    InterfaceIPv4.ADDRESS_PREFIX_LENGTH: 24,
                }
            ],
        },
    )

    with bridge.create() as state:
        assertlib.assert_state_match(state)

    assertlib.assert_absent(BRIDGE1)
    assertlib.assert_absent(PORT1)


@pytest.mark.tier1
def test_create_ovs_bridge_with_internal_port_with_copy_mac_from(eth1_up):
    eth1_mac = eth1_up[Interface.KEY][0][Interface.MAC]

    bridge = Bridge(BRIDGE1)
    bridge.add_internal_port(
        PORT1, copy_mac_from=eth1_up[Interface.KEY][0][Interface.NAME]
    )
    bridge.add_system_port(ETH1)

    with bridge.create():
        current_state = statelib.show_only((PORT1,))
        assertlib.assert_mac_address(current_state, eth1_mac)


@pytest.fixture
def ovs_bridge1_with_internal_port_same_name():
    bridge = Bridge(BRIDGE1)
    bridge.add_internal_port(
        BRIDGE1, ipv4_state={InterfaceIPv4.ENABLED: False}
    )

    with bridge.create() as state:
        yield state


@pytest.mark.tier1
def test_create_and_remove_ovs_bridge1_with_internal_port_same_name(
    ovs_bridge1_with_internal_port_same_name,
):
    state = statelib.show_only((BRIDGE1,))
    assert state
    assert len(state[Interface.KEY]) == 2


@pytest.mark.tier1
def test_vlan_as_ovs_bridge_port(vlan_on_eth1):
    bridge = Bridge(BRIDGE1)
    bridge.add_system_port(vlan_on_eth1)
    with bridge.create() as state:
        assertlib.assert_state_match(state)


@pytest.mark.tier1
def test_ovs_interface_with_max_length_name():
    bridge = Bridge(BRIDGE1)
    ovs_interface_name = "ovs123456789012"
    bridge.add_internal_port(ovs_interface_name)

    with bridge.create() as state:
        assertlib.assert_state_match(state)

    assertlib.assert_absent(BRIDGE1)
    assertlib.assert_absent(ovs_interface_name)


def test_nm_ovs_plugin_missing():
    with disable_nm_plugin("ovs"):
        with pytest.raises(NmstateDependencyError):
            libnmstate.apply(
                {
                    Interface.KEY: [
                        {
                            Interface.NAME: BRIDGE1,
                            Interface.TYPE: InterfaceType.OVS_BRIDGE,
                            Interface.STATE: InterfaceState.UP,
                        }
                    ]
                }
            )


@pytest.fixture
def vlan_on_eth1(eth1_up):
    with vlan_interface(
        VLAN_IFNAME, 101, eth1_up[Interface.KEY][0][Interface.NAME]
    ):
        yield VLAN_IFNAME


@pytest.mark.tier1
def test_change_ovs_interface_mac():
    bridge = Bridge(BRIDGE1)
    bridge.add_internal_port(PORT1, ipv4_state={InterfaceIPv4.ENABLED: False})

    with bridge.create() as state:
        desired_state = {
            Interface.KEY: [{Interface.NAME: PORT1, Interface.MAC: MAC1}]
        }
        libnmstate.apply(desired_state)
        assertlib.assert_state_match(state)

    assertlib.assert_absent(BRIDGE1)
    assertlib.assert_absent(PORT1)


class TestOvsLinkAggregation:
    @pytest.mark.tier1
    def test_create_and_remove_lag(self, port0_up, port1_up):
        port0_name = port0_up[Interface.KEY][0][Interface.NAME]
        port1_name = port1_up[Interface.KEY][0][Interface.NAME]

        bridge = Bridge(BRIDGE1)
        bridge.add_link_aggregation_port(BOND1, (port0_name, port1_name))

        with bridge.create() as state:
            assertlib.assert_state_match(state)

        assertlib.assert_absent(BRIDGE1)
        assertlib.assert_absent(BOND1)

    @pytest.mark.tier1
    def test_create_lag_with_ports_in_reverse_order(self, port0_up, port1_up):
        port0_name = port0_up[Interface.KEY][0][Interface.NAME]
        port1_name = port1_up[Interface.KEY][0][Interface.NAME]

        bridge = Bridge(BRIDGE1)
        bridge.add_link_aggregation_port(BOND1, (port1_name, port0_name))

        with bridge.create() as state:
            assertlib.assert_state_match(state)

        assertlib.assert_absent(BRIDGE1)
        assertlib.assert_absent(BOND1)

    def test_pretty_state_ovs_lag_name_first(self, eth1_up, eth2_up):
        bridge = Bridge(BRIDGE1)
        bridge.add_link_aggregation_port(
            BOND1,
            (ETH1, ETH2),
            mode=OVSBridge.Port.LinkAggregation.Mode.ACTIVE_BACKUP,
        )

        with bridge.create():
            current_state = statelib.show_only((BRIDGE1,))
            pretty_state = PrettyState(current_state)
            assert OVS_BOND_YAML_STATE in pretty_state.yaml

    @pytest.mark.tier1
    def test_add_ovs_lag_to_existing_ovs_bridge(self, port0_up, port1_up):
        with Bridge(BRIDGE1).create():
            port0_name = port0_up[Interface.KEY][0][Interface.NAME]
            port1_name = port1_up[Interface.KEY][0][Interface.NAME]
            bridge = Bridge(BRIDGE1)
            bridge.add_link_aggregation_port(BOND1, (port1_name, port0_name))
            libnmstate.apply(bridge.state)
            assertlib.assert_state_match(bridge.state)

    @pytest.mark.tier1
    def test_add_ovs_lag_with_updelay_and_downdelay(self, port0_up, port1_up):
        port0_name = port0_up[Interface.KEY][0][Interface.NAME]
        port1_name = port1_up[Interface.KEY][0][Interface.NAME]

        bridge = Bridge(BRIDGE1)
        bridge.add_link_aggregation_port(
            BOND1,
            (port0_name, port1_name),
            mode="active-backup",
            updelay=1000,
            downdelay=1000,
        )

        with bridge.create() as state:
            assertlib.assert_state_match(state)

        assertlib.assert_absent(BRIDGE1)
        assertlib.assert_absent(BOND1)

    @pytest.mark.tier1
    def test_modify_ovs_lag_with_updelay_and_downdelay(
        self, port0_up, port1_up
    ):
        port0_name = port0_up[Interface.KEY][0][Interface.NAME]
        port1_name = port1_up[Interface.KEY][0][Interface.NAME]

        bridge = Bridge(BRIDGE1)
        bridge.add_link_aggregation_port(
            BOND1,
            (port0_name, port1_name),
            mode="active-backup",
            updelay=1000,
            downdelay=1000,
        )

        with bridge.create() as state:
            assertlib.assert_state_match(state)
            state[Interface.KEY][0][OVSBridge.CONFIG_SUBTREE][
                OVSBridge.PORT_SUBTREE
            ][0][OVSBridge.Port.LINK_AGGREGATION_SUBTREE][
                OVSBridge.Port.LinkAggregation.Options.DOWN_DELAY
            ] = 100
            state[Interface.KEY][0][OVSBridge.CONFIG_SUBTREE][
                OVSBridge.PORT_SUBTREE
            ][0][OVSBridge.Port.LINK_AGGREGATION_SUBTREE][
                OVSBridge.Port.LinkAggregation.Options.UP_DELAY
            ] = 100
            libnmstate.apply(state)
            assertlib.assert_state_match(state)

        assertlib.assert_absent(BRIDGE1)
        assertlib.assert_absent(BOND1)


@pytest.mark.tier1
def test_ovs_vlan_access_tag():
    bridge = Bridge(BRIDGE1)
    bridge.add_internal_port(PORT1, ipv4_state={InterfaceIPv4.ENABLED: False})
    bridge.set_port_option(
        PORT1,
        {
            OVSBridge.Port.NAME: PORT1,
            OVSBridge.Port.VLAN_SUBTREE: {
                OVSBridge.Port.Vlan.MODE: OVSBridge.Port.Vlan.Mode.ACCESS,
                OVSBridge.Port.Vlan.TAG: 2,
            },
        },
    )
    with bridge.create() as state:
        assertlib.assert_state_match(state)

    assertlib.assert_absent(BRIDGE1)
    assertlib.assert_absent(PORT1)


@pytest.fixture
def bridge_port_with_trunks():
    bridge = Bridge(BRIDGE1)
    bridge.add_internal_port(PORT1, ipv4_state={InterfaceIPv4.ENABLED: False})
    bridge.set_port_option(
        PORT1,
        {
            OVSBridge.Port.NAME: PORT1,
            OVSBridge.Port.VLAN_SUBTREE: {
                OVSBridge.Port.Vlan.MODE: OVSBridge.Port.Vlan.Mode.TRUNK,
                OVSBridge.Port.Vlan.TAG: 0,
                OVSBridge.Port.Vlan.TRUNK_TAGS: [
                    {OVSBridge.Port.Vlan.TrunkTags.ID: 10},
                    {OVSBridge.Port.Vlan.TrunkTags.ID: 20},
                    {
                        OVSBridge.Port.Vlan.TrunkTags.ID_RANGE: {
                            OVSBridge.Port.Vlan.TrunkTags.MIN_RANGE: 30,
                            OVSBridge.Port.Vlan.TrunkTags.MAX_RANGE: 40,
                        }
                    },
                ],
            },
        },
    )
    with bridge.create() as state:
        yield state
    assertlib.assert_absent(BRIDGE1)
    assertlib.assert_absent(PORT1)


@pytest.mark.tier1
@pytest.mark.skipif(
    nm_minor_version() < 41,
    reason="OVS VLAN trunks was not supported in NM",
)
def test_ovs_vlan_trunks(bridge_port_with_trunks):
    assertlib.assert_state_match(bridge_port_with_trunks)


@pytest.mark.tier1
@pytest.mark.skipif(
    nm_minor_version() < 41,
    reason="OVS VLAN trunks was not supported in NM",
)
def test_remove_ovs_vlan_trunks(bridge_port_with_trunks):
    br1_state = statelib.show_only((BRIDGE1,))[Interface.KEY][0]
    port_state = br1_state[OVSBridge.CONFIG_SUBTREE][OVSBridge.PORT_SUBTREE][0]
    trunks = port_state[OVSBridge.Port.VLAN_SUBTREE][
        OVSBridge.Port.Vlan.TRUNK_TAGS
    ]
    assert len(trunks) == 3

    bridge_config = bridge_port_with_trunks[Interface.KEY][0][
        OVSBridge.CONFIG_SUBTREE
    ]
    port_config = bridge_config[OVSBridge.PORT_SUBTREE][0]
    port_config.update(
        {
            OVSBridge.Port.VLAN_SUBTREE: {
                OVSBridge.Port.Vlan.MODE: OVSBridge.Port.Vlan.Mode.TRUNK,
                OVSBridge.Port.Vlan.TAG: 0,
                OVSBridge.Port.Vlan.TRUNK_TAGS: [],
            }
        }
    )
    libnmstate.apply(bridge_port_with_trunks)
    assertlib.assert_state(bridge_port_with_trunks)

    br1_state = statelib.show_only((BRIDGE1,))[Interface.KEY][0]
    port_state = br1_state[OVSBridge.CONFIG_SUBTREE][OVSBridge.PORT_SUBTREE][0]
    trunks = port_state[OVSBridge.Port.VLAN_SUBTREE][
        OVSBridge.Port.Vlan.TRUNK_TAGS
    ]
    assert len(trunks) == 0


def test_add_invalid_port_ip_config(eth1_up):
    desired_state = eth1_up
    desired_state[Interface.KEY][0][Interface.IPV4][InterfaceIP.ENABLED] = True
    desired_state[Interface.KEY][0][Interface.IPV4][InterfaceIP.DHCP] = True
    with pytest.raises(NmstateValueError):
        bridge = Bridge(name=BRIDGE1)
        bridge.add_system_port("eth1")
        with bridge.create() as state:
            desired_state[Interface.KEY].append(state[Interface.KEY][0])
            libnmstate.apply(desired_state)


def test_ovsdb_new_bridge_with_external_id():
    bridge = Bridge(BRIDGE1)
    bridge.set_ovs_db({OvsDB.EXTERNAL_IDS: {"foo": "abc", "bak": 1}})
    bridge.add_internal_port(
        PORT1,
        ipv4_state={InterfaceIPv4.ENABLED: False},
        ovs_db={OvsDB.EXTERNAL_IDS: {"foo": "abcd", "bak": 2}},
    )
    with bridge.create() as state:
        assertlib.assert_state_match(state)
        new_state = statelib.show_only((PORT1,))
        # The newly created OVS internal interface should not hold
        # NM created external IDS.
        assert (
            "NM.connection.uuid"
            not in new_state[Interface.KEY][0][OvsDB.OVS_DB_SUBTREE][
                OvsDB.EXTERNAL_IDS
            ]
        )


def test_ovsdb_set_external_ids_for_ovs_system_interface(bridge_with_ports):
    system_port_name = "eth1"
    desired_state = {
        Interface.KEY: [
            {
                Interface.NAME: system_port_name,
                OvsDB.OVS_DB_SUBTREE: {OvsDB.EXTERNAL_IDS: {"foo": 1000}},
            }
        ]
    }
    libnmstate.apply(desired_state)
    assertlib.assert_state_match(desired_state)


def test_ovsdb_set_external_ids_for_existing_bridge(bridge_with_ports):
    bridge = bridge_with_ports
    bridge.set_ovs_db({OvsDB.EXTERNAL_IDS: {"foo": "abc", "bak": 1}})
    bridge.add_internal_port(
        PORT2,
        ipv4_state={InterfaceIPv4.ENABLED: False},
        ovs_db={OvsDB.EXTERNAL_IDS: {"foo": "abcd", "bak": 2}},
    )
    bridge.apply()
    assertlib.assert_state_match(bridge.state)


@pytest.fixture
def ovs_bridge_with_custom_external_ids():
    bridge = Bridge(BRIDGE1)
    bridge.add_internal_port(
        PORT1,
        ipv4_state={InterfaceIPv4.ENABLED: False},
        ovs_db={OvsDB.EXTERNAL_IDS: {"foo": "abcd", "bak": 2}},
    )
    with bridge.create() as state:
        yield state


def test_ovsdb_remove_external_ids(ovs_bridge_with_custom_external_ids):
    libnmstate.apply(
        {
            Interface.KEY: [
                {
                    Interface.NAME: PORT1,
                    OvsDB.OVS_DB_SUBTREE: {OvsDB.EXTERNAL_IDS: {}},
                }
            ]
        }
    )
    iface_info = statelib.show_only((PORT1,))[Interface.KEY][0]
    assert OvsDB.OVS_DB_SUBTREE not in iface_info


def test_ovsdb_override_external_ids(ovs_bridge_with_custom_external_ids):
    libnmstate.apply(
        {
            Interface.KEY: [
                {
                    Interface.NAME: PORT1,
                    OvsDB.OVS_DB_SUBTREE: {
                        OvsDB.EXTERNAL_IDS: {"new_ids": "haha"}
                    },
                }
            ]
        }
    )
    iface_info = statelib.show_only((PORT1,))[Interface.KEY][0]
    external_ids = iface_info[OvsDB.OVS_DB_SUBTREE][OvsDB.EXTERNAL_IDS]
    assert len(external_ids) == 1
    assert external_ids["new_ids"] == "haha"


def test_ovsdb_preserved_if_not_mentioned(ovs_bridge_with_custom_external_ids):
    libnmstate.apply(
        {Interface.KEY: [{Interface.NAME: PORT1, Interface.MTU: 1501}]}
    )
    iface_info = statelib.show_only((PORT1,))[Interface.KEY][0]
    external_ids = iface_info[OvsDB.OVS_DB_SUBTREE][OvsDB.EXTERNAL_IDS]
    assert len(external_ids) > 1


def test_ovsdb_global_config_add_external_ids_and_remove():
    desired_ovs_config = {
        OvsDB.EXTERNAL_IDS: {TEST_EXTERNAL_IDS_KEY: TEST_EXTERNAL_IDS_VALUE}
    }
    libnmstate.apply({OvsDB.KEY: desired_ovs_config})
    current_ovs_config = libnmstate.show()[OvsDB.KEY]

    assert state_match(desired_ovs_config, current_ovs_config)

    libnmstate.apply(
        {OvsDB.KEY: {OvsDB.EXTERNAL_IDS: {TEST_EXTERNAL_IDS_KEY: None}}}
    )
    current_ovs_config = libnmstate.show()[OvsDB.KEY]

    assert TEST_EXTERNAL_IDS_KEY not in current_ovs_config[OvsDB.EXTERNAL_IDS]


def test_ovsdb_global_config_add_other_config_and_remove():
    desired_ovs_config = {
        OvsDB.OTHER_CONFIG: {TEST_OTHER_CONFIG_KEY: TEST_OTHER_CONFIG_VALUE}
    }
    libnmstate.apply({OvsDB.KEY: desired_ovs_config})
    current_ovs_config = libnmstate.show()[OvsDB.KEY]

    assert state_match(desired_ovs_config, current_ovs_config)

    libnmstate.apply(
        {OvsDB.KEY: {OvsDB.OTHER_CONFIG: {TEST_OTHER_CONFIG_KEY: None}}}
    )
    current_ovs_config = libnmstate.show()[OvsDB.KEY]

    assert TEST_OTHER_CONFIG_KEY not in current_ovs_config[OvsDB.OTHER_CONFIG]


def test_remove_all_ovsdb_global_config():
    libnmstate.apply({OvsDB.KEY: {}})
    current_ovs_config = libnmstate.show()[OvsDB.KEY]

    assert current_ovs_config == {
        OvsDB.EXTERNAL_IDS: {},
        OvsDB.OTHER_CONFIG: {},
    }


@pytest.fixture
def ovsdb_global_config_external_ids():
    ovs_config = {
        OvsDB.EXTERNAL_IDS: {TEST_EXTERNAL_IDS_KEY: TEST_EXTERNAL_IDS_VALUE}
    }
    libnmstate.apply({OvsDB.KEY: ovs_config})
    yield ovs_config
    libnmstate.apply(
        {OvsDB.KEY: {OvsDB.EXTERNAL_IDS: {TEST_EXTERNAL_IDS_KEY: None}}}
    )


def test_ovsdb_global_config_untouched_if_not_defined(
    ovsdb_global_config_external_ids,
):
    desired_ovs_config = ovsdb_global_config_external_ids
    libnmstate.apply({})

    current_ovs_config = libnmstate.show()[OvsDB.KEY]
    assert state_match(desired_ovs_config, current_ovs_config)


@pytest.fixture
def ovs_bridge_with_patch_ports():
    patch0_state = {OVSInterface.Patch.PEER: "patch1"}
    patch1_state = {OVSInterface.Patch.PEER: "patch0"}
    bridge = Bridge(BRIDGE0)
    bridge.add_internal_port(PATCH0, patch_state=patch0_state)
    desired_state = bridge.state
    bridge = Bridge(BRIDGE1)
    bridge.add_internal_port(PATCH1, patch_state=patch1_state)
    desired_state[Interface.KEY].extend(bridge.state[Interface.KEY])
    libnmstate.apply(desired_state)
    assertlib.assert_state_match(desired_state)
    yield
    for iface in desired_state[Interface.KEY]:
        iface[Interface.STATE] = InterfaceState.ABSENT
    libnmstate.apply(desired_state)
    assertlib.assert_absent(BRIDGE1)
    assertlib.assert_absent(BRIDGE0)
    assertlib.assert_absent(PATCH0)
    assertlib.assert_absent(PATCH1)


class TestOvsPatch:
    def test_create_and_remove_patch_port(self, ovs_bridge_with_patch_ports):
        pass

    def test_patch_interface_does_not_have_mtu(self):
        patch0_state = {OVSInterface.Patch.PEER: "patch1"}
        patch1_state = {OVSInterface.Patch.PEER: "patch0"}
        bridge = Bridge(BRIDGE0)
        bridge.add_internal_port(PATCH0, patch_state=patch0_state)
        desired_state = bridge.state
        bridge = Bridge(BRIDGE1)
        bridge.add_internal_port(PATCH1, patch_state=patch1_state)
        desired_state[Interface.KEY].extend(bridge.state[Interface.KEY])
        try:
            libnmstate.apply(desired_state)
            patch0_state = statelib.show_only((PATCH0,))
            assert not patch0_state[Interface.KEY][0].get(Interface.MTU)
        finally:
            for iface in desired_state[Interface.KEY]:
                iface[Interface.STATE] = InterfaceState.ABSENT
            libnmstate.apply(desired_state)

    def test_add_patch_to_existing_interface_invalid(self):
        patch0_state = {OVSInterface.Patch.PEER: "falsepatch"}
        bridge = Bridge(BRIDGE0)
        bridge.add_internal_port(PATCH0)
        desired_state = bridge.state
        bridge = Bridge(BRIDGE1)
        bridge.add_internal_port(PATCH1)
        desired_state[Interface.KEY].extend(bridge.state[Interface.KEY])
        try:
            desired_state[Interface.KEY][1][
                OVSInterface.PATCH_CONFIG_SUBTREE
            ] = patch0_state
            desired_state[Interface.KEY][1][Interface.MTU] = 1500

            with pytest.raises(NmstateValueError):
                libnmstate.apply(desired_state)
        finally:
            for iface in desired_state[Interface.KEY]:
                iface[Interface.STATE] = InterfaceState.ABSENT
                iface[OVSInterface.PATCH_CONFIG_SUBTREE] = {}
            libnmstate.apply(desired_state)

        assertlib.assert_absent(BRIDGE1)
        assertlib.assert_absent(BRIDGE0)
        assertlib.assert_absent(PATCH0)
        assertlib.assert_absent(PATCH1)

    def test_add_patch_to_existing_interface_valid(self):
        patch0_state = {OVSInterface.Patch.PEER: "patch1"}
        patch1_state = {OVSInterface.Patch.PEER: "patch0"}
        bridge = Bridge(BRIDGE0)
        bridge.add_internal_port(PATCH0)
        desired_state = bridge.state
        bridge = Bridge(BRIDGE1)
        bridge.add_internal_port(PATCH1)
        desired_state[Interface.KEY].extend(bridge.state[Interface.KEY])
        try:
            desired_state[Interface.KEY][1].pop(Interface.MTU, None)
            desired_state[Interface.KEY][1].pop(Interface.MAC, None)
            desired_state[Interface.KEY][1][
                OVSInterface.PATCH_CONFIG_SUBTREE
            ] = patch0_state
            desired_state[Interface.KEY][3].pop(Interface.MTU, None)
            desired_state[Interface.KEY][3].pop(Interface.MAC, None)
            desired_state[Interface.KEY][3][
                OVSInterface.PATCH_CONFIG_SUBTREE
            ] = patch1_state

            libnmstate.apply(desired_state)
            assertlib.assert_state_match(desired_state)
        finally:
            for iface in desired_state[Interface.KEY]:
                iface[Interface.STATE] = InterfaceState.ABSENT
            libnmstate.apply(desired_state)

        assertlib.assert_absent(BRIDGE1)
        assertlib.assert_absent(BRIDGE0)
        assertlib.assert_absent(PATCH0)
        assertlib.assert_absent(PATCH1)

    def test_change_ovsdb_ext_id_of_ovs_path(
        self, ovs_bridge_with_patch_ports
    ):
        libnmstate.apply(
            {
                Interface.KEY: [
                    {
                        Interface.NAME: PATCH0,
                        OvsDB.KEY: {
                            "foo": "abc",
                        },
                    },
                    {
                        Interface.NAME: PATCH1,
                        OvsDB.KEY: {
                            "foo": "abd",
                        },
                    },
                ]
            }
        )


@pytest.mark.tier1
def test_create_internal_port_with_explict_mtu():
    bridge = Bridge(BRIDGE1)
    bridge.add_internal_port(
        PORT1,
        mac=MAC1,
        mtu=1501,
        ipv4_state={
            InterfaceIPv4.ENABLED: True,
            InterfaceIPv4.ADDRESS: [
                {
                    InterfaceIPv4.ADDRESS_IP: "192.0.2.1",
                    InterfaceIPv4.ADDRESS_PREFIX_LENGTH: 24,
                }
            ],
        },
    )

    with bridge.create() as state:
        assertlib.assert_state_match(state)

    assertlib.assert_absent(BRIDGE1)
    assertlib.assert_absent(PORT1)


@pytest.mark.tier1
def test_change_mtu_of_existing_internal_port(bridge_with_ports):
    state = statelib.show_only((PORT1,))
    state[Interface.KEY][0][Interface.MTU] = 1501

    libnmstate.apply(state)

    assertlib.assert_state_match(state)


@pytest.mark.tier1
def test_create_ovs_with_internal_ports_in_reverse_order():
    bridge = Bridge(BRIDGE1)
    bridge.add_internal_port(PORT2)
    bridge.add_internal_port(PORT1)

    with bridge.create() as state:
        assertlib.assert_state_match(state)

    assertlib.assert_absent(BRIDGE1)
    assertlib.assert_absent(PORT1)
    assertlib.assert_absent(PORT2)


def test_create_memory_only_ovs_bridge():
    bridge = Bridge(BRIDGE1)
    bridge.add_internal_port(PORT1)

    if nm_major_minor_version() <= 1.28:
        with pytest.raises(NmstateNotSupportedError):
            libnmstate.apply(bridge.state, save_to_disk=False)
    else:
        try:
            libnmstate.apply(bridge.state, save_to_disk=False)
        finally:
            libnmstate.apply(
                {
                    Interface.KEY: [
                        {
                            Interface.NAME: BRIDGE1,
                            Interface.STATE: InterfaceState.ABSENT,
                        }
                    ]
                }
            )


def test_remove_all_ovs_ports(bridge_with_ports):
    libnmstate.apply(
        {
            Interface.KEY: [
                {
                    Interface.NAME: BRIDGE1,
                    OVSBridge.CONFIG_SUBTREE: {OVSBridge.PORT_SUBTREE: []},
                }
            ]
        }
    )
    assertlib.assert_absent(PORT1)


def test_modify_state_do_not_remove_unmanaged_ovs(eth1_up):
    with unmanaged_ovs_bridge():
        eth1_up[Interface.KEY][0][Interface.MTU] = 1400
        libnmstate.apply(eth1_up)
        current_state = statelib.show_only((BRIDGE0,))
        assert len(current_state[Interface.KEY])
        assert current_state[Interface.KEY][0][Interface.NAME] == BRIDGE0


@contextmanager
def unmanaged_ovs_bridge():
    rc, _, _ = cmdlib.exec_cmd(
        "ovs-vsctl add-br br0 -- set Bridge br0 fail-mode=secure".split(),
        check=True,
    )
    assert rc == RC_SUCCESS
    try:
        yield
    finally:
        rc, _, _ = cmdlib.exec_cmd("ovs-vsctl del-br br0".split(), check=True)
        assert rc == RC_SUCCESS


def test_expect_failure_when_create_ovs_interface_without_bridge():
    with pytest.raises(NmstateValueError):
        libnmstate.apply(
            {
                Interface.KEY: [
                    {
                        Interface.NAME: "ovs0",
                        Interface.TYPE: InterfaceType.OVS_INTERFACE,
                        Interface.STATE: InterfaceState.UP,
                    }
                ]
            }
        )


@pytest.mark.tier1
def test_create_vlan_over_ovs_iface_with_use_same_name_as_bridge(
    ovs_bridge1_with_internal_port_same_name,
):
    desired_state = {
        Interface.KEY: [
            {
                Interface.NAME: "vlan101",
                Interface.TYPE: InterfaceType.VLAN,
                VLAN.CONFIG_SUBTREE: {VLAN.ID: 101, VLAN.BASE_IFACE: BRIDGE1},
            }
        ]
    }
    libnmstate.apply(desired_state)
    assertlib.assert_state_match(desired_state)


def test_create_vxlan_over_ovs_iface_with_use_same_name_as_bridge(
    ovs_bridge1_with_internal_port_same_name,
):
    desired_state = {
        Interface.KEY: [
            {
                Interface.NAME: "vlan101",
                Interface.TYPE: InterfaceType.VXLAN,
                VXLAN.CONFIG_SUBTREE: {
                    VXLAN.ID: 101,
                    VXLAN.BASE_IFACE: BRIDGE1,
                    VXLAN.REMOTE: "192.0.2.251",
                },
            }
        ]
    }
    libnmstate.apply(desired_state)
    assertlib.assert_state_match(desired_state)


def test_create_mac_vlan_over_ovs_iface_with_use_same_name_as_bridge(
    ovs_bridge1_with_internal_port_same_name,
):
    desired_state = {
        Interface.KEY: [
            {
                Interface.NAME: "mac_vlan101",
                Interface.TYPE: InterfaceType.MAC_VLAN,
                MacVlan.CONFIG_SUBTREE: {
                    MacVlan.BASE_IFACE: BRIDGE1,
                    MacVlan.MODE: MacVlan.Mode.PASSTHRU,
                    MacVlan.PROMISCUOUS: False,
                },
            }
        ]
    }
    libnmstate.apply(desired_state)
    assertlib.assert_state_match(desired_state)


def test_create_mac_tap_over_ovs_iface_with_use_same_name_as_bridge(
    ovs_bridge1_with_internal_port_same_name,
):
    desired_state = {
        Interface.KEY: [
            {
                Interface.NAME: "mac_tap0",
                Interface.TYPE: InterfaceType.MAC_VTAP,
                MacVtap.CONFIG_SUBTREE: {
                    MacVtap.BASE_IFACE: BRIDGE1,
                    MacVtap.MODE: MacVtap.Mode.PASSTHRU,
                    MacVtap.PROMISCUOUS: False,
                },
            }
        ]
    }
    libnmstate.apply(desired_state)
    assertlib.assert_state_match(desired_state)


def test_ignore_ovs_system_kernel_nic(bridge_with_ports):
    assert not statelib.show_only(("ovs-system",))[Interface.KEY]


@pytest.mark.tier1
def test_set_static_to_ovs_interface_with_the_same_name_bridge(
    ovs_bridge1_with_internal_port_same_name,
):
    desired_state = {
        Interface.KEY: [
            {
                Interface.NAME: BRIDGE1,
                Interface.TYPE: InterfaceType.OVS_INTERFACE,
                Interface.IPV4: {
                    InterfaceIPv4.ENABLED: True,
                    InterfaceIPv4.ADDRESS: [
                        {
                            InterfaceIPv4.ADDRESS_IP: "192.0.2.1",
                            InterfaceIPv4.ADDRESS_PREFIX_LENGTH: 24,
                        }
                    ],
                },
            }
        ]
    }
    libnmstate.apply(desired_state)

    def _ovs_iface_got_ip():
        state = statelib.show_only((BRIDGE1,))
        if len(state[Interface.KEY]) != 2:
            return False
        cur_iface_state = None
        for iface in state[Interface.KEY]:
            if iface[Interface.TYPE] == InterfaceType.OVS_INTERFACE:
                cur_iface_state = iface
                break
        return cur_iface_state and cur_iface_state[Interface.IPV4][
            InterfaceIPv4.ADDRESS
        ] == [
            {
                InterfaceIPv4.ADDRESS_IP: "192.0.2.1",
                InterfaceIPv4.ADDRESS_PREFIX_LENGTH: 24,
            }
        ]

    retry_till_true_or_timeout(RETRY_TIMEOUT, _ovs_iface_got_ip)


@pytest.fixture
def ovs_bridge_with_dhcp_route_table_500(bridge_with_ports):
    desired_state = {
        Interface.KEY: [
            {
                Interface.NAME: PORT1,
                Interface.TYPE: InterfaceType.OVS_INTERFACE,
                Interface.IPV4: {
                    InterfaceIPv4.ENABLED: True,
                    InterfaceIPv4.DHCP: True,
                    InterfaceIPv4.AUTO_ROUTE_TABLE_ID: 500,
                },
            }
        ]
    }
    libnmstate.apply(desired_state)
    yield


@pytest.mark.tier1
def test_add_route_rule_to_ovs_interface_dhcp_auto_route_table(
    ovs_bridge_with_dhcp_route_table_500,
):
    route_rule = {
        RouteRule.IP_FROM: "192.0.2.0/24",
        RouteRule.ROUTE_TABLE: 500,
    }
    desired_state = {RouteRule.KEY: {RouteRule.CONFIG: [route_rule]}}
    libnmstate.apply(desired_state)

    iprule.ip_rule_exist_in_os(
        route_rule.get(RouteRule.IP_FROM),
        route_rule.get(RouteRule.IP_TO),
        route_rule.get(RouteRule.PRIORITY),
        route_rule.get(RouteRule.ROUTE_TABLE),
        route_rule.get(RouteRule.FWMARK),
        route_rule.get(RouteRule.FWMASK),
        route_rule.get(RouteRule.FAMILY),
    )


@pytest.fixture
def cleanup_ovs_bridge_and_bond():
    yield
    libnmstate.apply(
        {
            Interface.KEY: [
                {
                    Interface.NAME: BRIDGE0,
                    Interface.STATE: InterfaceState.ABSENT,
                },
                {
                    Interface.NAME: BOND1,
                    Interface.STATE: InterfaceState.ABSENT,
                },
            ]
        }
    )


@pytest.mark.tier1
def test_attach_linux_bond_to_ovs_bridge(
    cleanup_ovs_bridge_and_bond, eth1_up, eth2_up
):
    desired_state = yaml.load(
        """---
        interfaces:
          - name: bond1
            state: up
            type: bond
            link-aggregation:
              mode: active-backup
              options:
                miimon: 140
                primary: eth1
              port:
                - eth1
                - eth2
          - name: br0
            type: ovs-bridge
            state: up
            bridge:
              options:
                stp: false
              port:
                - name: bond1
            """,
        Loader=yaml.SafeLoader,
    )
    libnmstate.apply(desired_state)
    assertlib.assert_state_match(desired_state)


@pytest.mark.skipif(
    not os.environ.get("TEST_PCI_PATH"),
    reason="Need to define TEST_PCI_PATH for OVS DPDK tests.",
)
class TestOvsDpdk:
    def test_create_ovs_dpdk_and_remove(self):
        dpdk0_state = {OVSInterface.Dpdk.DEVARGS: _test_pci_path()}
        bridge = Bridge(BRIDGE0)
        bridge.add_internal_port(PORT1, dpdk_state=dpdk0_state)
        bridge.set_options({OVSBridge.Options.DATAPATH: "netdev"})
        desired_state = bridge.state
        try:
            libnmstate.apply(desired_state)
            assertlib.assert_state_match(desired_state)
        finally:
            for iface in desired_state[Interface.KEY]:
                iface[Interface.STATE] = InterfaceState.ABSENT
            libnmstate.apply(desired_state)

        assertlib.assert_absent(BRIDGE0)
        assertlib.assert_absent(PORT1)

    @pytest.mark.tier1
    def test_create_ovs_dpdk_with_rx_queue(self):
        dpdk0_state = {
            OVSInterface.Dpdk.DEVARGS: _test_pci_path(),
            OVSInterface.Dpdk.RX_QUEUE: 10,
        }
        bridge = Bridge(BRIDGE0)
        bridge.add_internal_port(PORT1, dpdk_state=dpdk0_state)
        bridge.set_options({OVSBridge.Options.DATAPATH: "netdev"})
        desired_state = bridge.state
        try:
            libnmstate.apply(desired_state)
            assertlib.assert_state_match(desired_state)
        finally:
            for iface in desired_state[Interface.KEY]:
                iface[Interface.STATE] = InterfaceState.ABSENT
            libnmstate.apply(desired_state)

        assertlib.assert_absent(BRIDGE0)
        assertlib.assert_absent(PORT1)

    @pytest.mark.parametrize("datapaths", ("netdev", "system"))
    def test_create_ovs_dpdk_with_datapath(self, datapaths):
        dpdk0_state = {OVSInterface.Dpdk.DEVARGS: _test_pci_path()}
        bridge = Bridge(BRIDGE0)
        bridge.add_internal_port(PORT1, dpdk_state=dpdk0_state)
        bridge.set_options({OVSBridge.Options.DATAPATH: datapaths})
        desired_state = bridge.state
        try:
            libnmstate.apply(desired_state)
            assertlib.assert_state_match(desired_state)
        finally:
            for iface in desired_state[Interface.KEY]:
                iface[Interface.STATE] = InterfaceState.ABSENT
            libnmstate.apply(desired_state)

        assertlib.assert_absent(BRIDGE0)
        assertlib.assert_absent(PORT1)

    def test_create_ovs_dpdk_queue_descriptor(self, datapaths):
        dpdk0_state = {
            OVSInterface.Dpdk.DEVARGS: _test_pci_path(),
            OVSInterface.Dpdk.N_RXQ_DESC: 1024,
            OVSInterface.Dpdk.N_TXQ_DESC: 2048,
        }
        bridge = Bridge(BRIDGE0)
        bridge.add_internal_port(PORT1, dpdk_state=dpdk0_state)
        bridge.set_options({OVSBridge.Options.DATAPATH: "netdev"})
        desired_state = bridge.state
        try:
            libnmstate.apply(desired_state)
            assertlib.assert_state_match(desired_state)
        finally:
            for iface in desired_state[Interface.KEY]:
                iface[Interface.STATE] = InterfaceState.ABSENT
            libnmstate.apply(desired_state)

        assertlib.assert_absent(BRIDGE0)
        assertlib.assert_absent(PORT1)


@pytest.fixture
def unmanged_ovs_vxlan():
    cmdlib.exec_cmd("ovs-vsctl add-br br0_with_vxlan".split(), check=True)
    cmdlib.exec_cmd(
        "ovs-vsctl add-port br0_with_vxlan vx_node1 -- "
        "set interface vx_node1 type=vxlan options:remote_ip=192.168.122.174 "
        "options:dst_port=8472".split(),
        check=True,
    )
    cmdlib.exec_cmd(
        "nmcli d set vxlan_sys_8472 managed false".split(), check=True
    )
    try:
        yield
    finally:
        cmdlib.exec_cmd("ovs-vsctl del-br br0_with_vxlan".split())


@pytest.mark.tier1
def test_ovs_vxlan_in_current_not_impact_others(unmanged_ovs_vxlan):
    bridge = Bridge(BRIDGE1)
    bridge.add_internal_port(PORT1, ipv4_state={InterfaceIPv4.ENABLED: False})
    with bridge.create() as state:
        assertlib.assert_state_match(state)

    assertlib.assert_absent(BRIDGE1)


def test_add_new_ovs_interface_to_existing(bridge_with_ports):
    bridge = bridge_with_ports
    bridge.add_internal_port(
        PORT2,
        ipv4_state={InterfaceIPv4.ENABLED: False},
    )
    bridge.apply()
    assertlib.assert_state_match(bridge.state)


@pytest.fixture
def clean_new_veth():
    yield
    libnmstate.apply(
        {
            Interface.KEY: [
                {Interface.NAME: VETH1, Interface.STATE: InterfaceState.ABSENT}
            ]
        }
    )


def test_add_new_sys_veth_interface_to_existing_ovs_bridge(
    bridge_with_ports, clean_new_veth
):
    bridge = bridge_with_ports
    bridge.add_system_port(VETH1)
    desired_state = bridge.state
    desired_state[Interface.KEY].append(
        {
            Interface.NAME: VETH1,
            Interface.TYPE: InterfaceType.VETH,
            Veth.CONFIG_SUBTREE: {
                Veth.PEER: VETH1_PEER,
            },
        }
    )
    libnmstate.apply(desired_state)
    assertlib.assert_state_match(desired_state)


@pytest.fixture
def ovsdb_global_db():
    original = libnmstate.show()[OvsDB.KEY]
    libnmstate.apply(
        {
            OvsDB.KEY: {
                OvsDB.EXTERNAL_IDS: {
                    "opt1": "value1",
                    "opt2": "value2",
                },
                OvsDB.OTHER_CONFIG: {
                    "flow-restore-wait": "true",
                },
            }
        }
    )
    yield
    libnmstate.apply({OvsDB.KEY: {}})
    libnmstate.apply({OvsDB.KEY: original})


def test_ovsdb_global_merged_desired_with_current(ovsdb_global_db):
    desired_state = {
        OvsDB.KEY: {
            OvsDB.EXTERNAL_IDS: {
                "opt1": None,
                "opt2": "value2new",
            },
            OvsDB.OTHER_CONFIG: {
                "flow-restore-wait": None,
                "tc-policy": "skip_hw",
            },
        }
    }
    libnmstate.apply(desired_state)

    current = libnmstate.show()[OvsDB.KEY]

    assert "opt1" not in current[OvsDB.EXTERNAL_IDS]
    assert current[OvsDB.EXTERNAL_IDS]["opt2"] == "value2new"
    assert "flow-restore-wait" not in current[OvsDB.OTHER_CONFIG]
    assert current[OvsDB.OTHER_CONFIG]["tc-policy"] == "skip_hw"


def test_ovsdb_global_preserve_not_mentioned(ovsdb_global_db):
    desired_state = {
        OvsDB.KEY: {
            OvsDB.OTHER_CONFIG: {
                "flow-restore-wait": None,
                "tc-policy": "skip_hw",
            },
        }
    }
    libnmstate.apply(desired_state)

    current = libnmstate.show()[OvsDB.KEY]

    assert current[OvsDB.EXTERNAL_IDS]["opt1"] == "value1"
    assert current[OvsDB.EXTERNAL_IDS]["opt2"] == "value2"
    assert "flow-restore-wait" not in current[OvsDB.OTHER_CONFIG]
    assert current[OvsDB.OTHER_CONFIG]["tc-policy"] == "skip_hw"


def test_ovsdb_global_remove_all(ovsdb_global_db):
    desired_state = {OvsDB.KEY: {}}
    libnmstate.apply(desired_state)

    current = libnmstate.show()[OvsDB.KEY]

    assert not current[OvsDB.EXTERNAL_IDS]


def test_ovsdb_global_remove_all_external_ids(ovsdb_global_db):
    desired_state = {OvsDB.KEY: {OvsDB.EXTERNAL_IDS: {}}}
    libnmstate.apply(desired_state)

    current = libnmstate.show()[OvsDB.KEY]

    assert not current[OvsDB.EXTERNAL_IDS]
    assert current[OvsDB.OTHER_CONFIG]["flow-restore-wait"] == "true"


def test_ovsdb_global_remove_all_other_config(ovsdb_global_db):
    desired_state = {OvsDB.KEY: {OvsDB.OTHER_CONFIG: {}}}
    libnmstate.apply(desired_state)

    current = libnmstate.show()[OvsDB.KEY]

    assert not current[OvsDB.OTHER_CONFIG]
    assert current[OvsDB.EXTERNAL_IDS]["opt1"] == "value1"
    assert current[OvsDB.EXTERNAL_IDS]["opt2"] == "value2"


def test_change_ovs_intenral_iface_ext_id_with_br_port_not_mentioned(
    bridge_with_ports,
):
    desired_state = {
        Interface.KEY: [
            {
                Interface.NAME: BRIDGE1,
                Interface.TYPE: InterfaceType.OVS_BRIDGE,
            },
            {
                Interface.NAME: PORT1,
                Interface.TYPE: InterfaceType.OVS_INTERFACE,
                OvsDB.OVS_DB_SUBTREE: {
                    OvsDB.EXTERNAL_IDS: {"foo": "abc", "bak": 1}
                },
            },
        ]
    }
    libnmstate.apply(desired_state)
    assertlib.assert_state_match(desired_state)


def test_move_ovs_system_interface_to_bond(bridge_with_ports):
    with bond_interface(BOND1, ["eth1"], create=False) as desired_state:
        desired_state[Interface.KEY].append(
            {
                Interface.NAME: BRIDGE1,
                OVSBridge.CONFIG_SUBTREE: {
                    OVSBridge.PORT_SUBTREE: [
                        {
                            OVSBridge.Port.NAME: PORT1,
                        }
                    ]
                },
            }
        )
        libnmstate.apply(desired_state)


@pytest.mark.skipif(is_k8s(), reason="K8S does not support genconf")
def test_genconf_ovsdb_iface_external_ids(eth1_up):
    desired_state = {
        Interface.KEY: [
            {
                Interface.NAME: BRIDGE1,
                Interface.TYPE: InterfaceType.OVS_BRIDGE,
                OVSBridge.CONFIG_SUBTREE: {
                    OVSBridge.PORT_SUBTREE: [
                        {
                            OVSBridge.Port.NAME: PORT1,
                        },
                        {
                            OVSBridge.Port.NAME: "eth1",
                        },
                    ]
                },
                OvsDB.OVS_DB_SUBTREE: {
                    OvsDB.EXTERNAL_IDS: {"foo": "abe", "bak": 3}
                },
            },
            {
                Interface.NAME: "eth1",
                Interface.TYPE: InterfaceType.ETHERNET,
                OvsDB.OVS_DB_SUBTREE: {
                    OvsDB.EXTERNAL_IDS: {"foo": "abd", "bak": 2}
                },
            },
            {
                Interface.NAME: PORT1,
                Interface.TYPE: InterfaceType.OVS_INTERFACE,
                OvsDB.OVS_DB_SUBTREE: {
                    OvsDB.EXTERNAL_IDS: {"foo": "abc", "bak": 1}
                },
            },
        ]
    }
    with gen_conf_apply(desired_state):
        assertlib.assert_state_match(desired_state)


def test_add_port_to_ovs_br_with_controller_property(
    ovs_bridge1_with_internal_port_same_name, eth2_up
):
    libnmstate.apply(
        {
            Interface.KEY: [
                {
                    Interface.NAME: "eth2",
                    Interface.STATE: InterfaceState.UP,
                    Interface.CONTROLLER: BRIDGE1,
                }
            ]
        }
    )
    current_state = libnmstate.show()
    br_iface = None
    # The show_only() does not works well with ovs same name topology, hence
    # we use our own code.
    for iface in current_state[Interface.KEY]:
        if (
            iface[Interface.NAME] == BRIDGE1
            and iface[Interface.TYPE] == InterfaceType.OVS_BRIDGE
        ):
            br_iface = iface
            break
    br_ports = br_iface[OVSBridge.CONFIG_SUBTREE][OVSBridge.PORT_SUBTREE]
    assert br_iface[Interface.NAME] == BRIDGE1
    assert len(br_ports) == 2
    assert br_ports[0][OVSBridge.Port.NAME] == BRIDGE1
    assert br_ports[1][OVSBridge.Port.NAME] == "eth2"


def test_move_ovs_sys_iface_to_linux_bridge(bridge_with_ports):
    bridge_config = {
        LinuxBridge.OPTIONS_SUBTREE: {
            LinuxBridge.STP_SUBTREE: {
                LinuxBridge.STP.ENABLED: False,
            },
        },
        LinuxBridge.PORT_SUBTREE: [
            {
                LinuxBridge.Port.NAME: "eth1",
            }
        ],
    }
    with linux_bridge("test-linux-br0", bridge_config) as state:
        assertlib.assert_state_match(state)


def test_ovs_vlan_access_mode_with_tag_0():
    bridge = Bridge(BRIDGE1)
    bridge.add_internal_port(PORT1, ipv4_state={InterfaceIPv4.ENABLED: False})
    bridge.set_port_option(
        PORT1,
        {
            OVSBridge.Port.NAME: PORT1,
            OVSBridge.Port.VLAN_SUBTREE: {
                OVSBridge.Port.Vlan.MODE: OVSBridge.Port.Vlan.Mode.ACCESS,
                OVSBridge.Port.Vlan.TAG: 0,
            },
        },
    )
    with bridge.create() as state:
        assertlib.assert_state_match(state)

    assertlib.assert_absent(BRIDGE1)
    assertlib.assert_absent(PORT1)


@pytest.fixture
def cleanup_ovs_bridge():
    yield
    libnmstate.apply(
        {
            Interface.KEY: [
                {
                    Interface.NAME: BRIDGE0,
                    Interface.STATE: InterfaceState.ABSENT,
                },
            ]
        }
    )


@pytest.mark.parametrize(
    "bond_mode",
    [
        OVSBridge.Port.LinkAggregation.Mode.ACTIVE_BACKUP,
        OVSBridge.Port.LinkAggregation.Mode.BALANCE_SLB,
        OVSBridge.Port.LinkAggregation.Mode.BALANCE_TCP,
        OVSBridge.Port.LinkAggregation.Mode.LACP,
    ],
    ids=["active-backup", "balance-slb", "balance-tcp", "lacp"],
)
def test_crate_ovs_bond(cleanup_ovs_bridge, eth1_up, eth2_up, bond_mode):
    desired_state = yaml.load(
        """---
        interfaces:
        - name: br0
          type: ovs-bridge
          state: up
          bridge:
            port:
            - name: ovs0
            - name: bond0
              link-aggregation:
                mode: {bond_mode}
                port:
                - name: eth1
                - name: eth2
            """.format(
            bond_mode=bond_mode
        ),
        Loader=yaml.SafeLoader,
    )
    libnmstate.apply(desired_state)
    assertlib.assert_state_match(desired_state)


@pytest.fixture
def ovs_service_off():
    with disable_service("openvswitch"):
        yield


def test_global_ovsdb_with_ovs_service_off(ovs_service_off):
    libnmstate.apply({OvsDB.KEY: {}})


@pytest.fixture
def ovs_br_with_4_dummy_ports_ovs_bond():
    desired_state = yaml.load(
        """---
        interfaces:
        - name: br0
          type: ovs-bridge
          state: up
          bridge:
            port:
            - name: ovs-bond0
              link-aggregation:
                mode: balance-slb
                port:
                - name: dummy4
                - name: dummy1
                - name: dummy3
                - name: dummy2
        - name: dummy1
          type: dummy
          state: up
        - name: dummy2
          type: dummy
          state: up
        - name: dummy3
          type: dummy
          state: up
        - name: dummy4
          type: dummy
          state: up
        """,
        Loader=yaml.SafeLoader,
    )
    libnmstate.apply(desired_state)
    yield
    desired_state = yaml.load(
        """---
        interfaces:
        - name: br0
          type: ovs-bridge
          state: absent
        - name: dummy1
          type: dummy
          state: absent
        - name: dummy2
          type: dummy
          state: absent
        - name: dummy3
          type: dummy
          state: absent
        - name: dummy4
          type: dummy
          state: absent
        """,
        Loader=yaml.SafeLoader,
    )
    libnmstate.apply(desired_state)


def test_ovs_detach_2_ports_from_4_ports_ovs_bond(
    ovs_br_with_4_dummy_ports_ovs_bond,
):
    desired_state = yaml.load(
        """---
        interfaces:
        - name: br0
          type: ovs-bridge
          state: up
          bridge:
            port:
            - name: ovs-bond0
              link-aggregation:
                mode: balance-slb
                port:
                - name: dummy2
                - name: dummy4
        """,
        Loader=yaml.SafeLoader,
    )
    libnmstate.apply(desired_state)
    assertlib.assert_state_match(desired_state)


# OpenStack use case
@pytest.mark.tier1
@pytest.mark.skipif(
    nm_minor_version() < 41,
    reason="OVS interface level other_config is not supported in NM 1.40-",
)
def test_ovs_bond_other_config_and_remove(
    cleanup_ovs_bridge, eth1_up, eth2_up
):
    desired_state = yaml.load(
        """---
        interfaces:
        - name: br0
          type: ovs-bridge
          state: up
          bridge:
            port:
            - name: ovs0
            - name: bond0
              link-aggregation:
                mode: balance-slb
                port:
                - name: eth1
                - name: eth2
                ovs-db:
                  external_ids:
                    test_str: foo1
                    test_num: 100
                  other_config:
                    bond-miimon-interval: 100
        """,
        Loader=yaml.SafeLoader,
    )
    libnmstate.apply(desired_state)
    assertlib.assert_state_match(desired_state)

    desired_state = yaml.load(
        """---
        interfaces:
        - name: br0
          type: ovs-bridge
          state: up
          bridge:
            port:
            - name: ovs0
            - name: bond0
              link-aggregation:
                mode: balance-slb
                port:
                - name: eth1
                - name: eth2
                ovs-db: {}
        """,
        Loader=yaml.SafeLoader,
    )
    libnmstate.apply(desired_state)
    assertlib.assert_state_match(desired_state)


# OpenStack use case
@pytest.mark.tier1
@pytest.mark.skipif(
    nm_minor_version() < 41,
    reason="OVS interface level other_config is not supported in NM 1.40-",
)
def test_ovs_bridge_other_config_and_remove(
    cleanup_ovs_bridge, eth1_up, eth2_up
):
    desired_state = yaml.load(
        """---
        interfaces:
        - name: br0
          type: ovs-bridge
          state: up
          ovs-db:
            other_config:
              in-band-queue: 12
          bridge:
            port:
            - name: ovs0
            - name: bond0
              link-aggregation:
                mode: balance-slb
                port:
                - name: eth1
                - name: eth2
        """,
        Loader=yaml.SafeLoader,
    )
    libnmstate.apply(desired_state)
    assertlib.assert_state_match(desired_state)
    desired_state = yaml.load(
        """---
        interfaces:
        - name: br0
          type: ovs-bridge
          state: up
          ovs-db: {}
          bridge:
            port:
            - name: ovs0
            - name: bond0
              link-aggregation:
                mode: balance-slb
                port:
                - name: eth1
                - name: eth2
        """,
        Loader=yaml.SafeLoader,
    )
    libnmstate.apply(desired_state)
    assertlib.assert_state_match(desired_state)


# OpenStack use case
@pytest.mark.tier1
@pytest.mark.skipif(
    nm_minor_version() < 41,
    reason="OVS interface level other_config is not supported in NM 1.40-",
)
def test_ovs_sys_iface_other_config_and_remove(
    cleanup_ovs_bridge, eth1_up, eth2_up
):
    desired_state = yaml.load(
        """---
        interfaces:
        - name: eth1
          type: ethernet
          state: up
          ovs-db:
            other_config:
              emc-insert-inv-prob: 90
        - name: br0
          type: ovs-bridge
          state: up
          bridge:
            port:
            - name: ovs0
            - name: eth1
        """,
        Loader=yaml.SafeLoader,
    )
    libnmstate.apply(desired_state)
    assertlib.assert_state_match(desired_state)

    desired_state = yaml.load(
        """---
        interfaces:
        - name: eth1
          type: ethernet
          state: up
          ovs-db: {}
        """,
        Loader=yaml.SafeLoader,
    )
    libnmstate.apply(desired_state)
    assertlib.assert_state_match(desired_state)


@pytest.fixture
def ovs_bridge_with_auto_create_internal_iface():
    with Bridge(BRIDGE1).create():
        yield


def test_ovs_new_internal_iface_to_bridge_with_auto_create_iface(
    ovs_bridge_with_auto_create_internal_iface,
):
    desired_state = yaml.load(
        """---
        interfaces:
        - name: br1
          type: ovs-bridge
          state: up
          bridge:
            port:
            - name: new_ovs0
        """,
        Loader=yaml.SafeLoader,
    )
    libnmstate.apply(desired_state)
    assertlib.assert_state_match(desired_state)


def test_ovs_replace_internal_iface_to_bridge_with_auto_create_iface(
    ovs_bridge_with_auto_create_internal_iface,
):
    desired_state = yaml.load(
        """---
        interfaces:
        - name: br1
          type: ovs-interface
          state: absent
        - name: br1
          type: ovs-bridge
          state: up
          bridge:
            port:
            - name: ovs0
        """,
        Loader=yaml.SafeLoader,
    )
    libnmstate.apply(desired_state)
    assertlib.assert_state_match(desired_state)
